#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import sys
from collections.abc import Mapping, Sequence
from functools import lru_cache
from inspect import Parameter, formatannotation, isclass, signature
from types import NoneType, UnionType
from typing import (
    Any,
    Generic,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from xrlint.util.formatting import format_message_one_of, format_message_type_of

T = TypeVar("T")

_IS_PYTHON_3_10 = (3, 10) <= sys.version_info < (3, 11)


class ValueConstructible(Generic[T]):
    """A mixin that makes your classes constructible from a single value
    of any type.

    Implementing classes override one of the many `_from_<type>()`
    class methods to implement support converting from values of
    type `<type>`. They may use the [_from_typed_value][] to convert values
    from values with given type annotations, such as object properties
    or constructor parameters.

    The factory for this purpose is the
    class method [from_value][xrlint.util.codec.ValueConstructible.from_value].
    """

    @classmethod
    def from_value(cls, value: Any, value_name: str | None = None) -> T:
        """Create an instance of this class from a value.

        The default implementation checks if `value` is already an
        instance of this class. If so, it is returned unchanged.

        It then delegates to various `_from_<type>()` methods which
        all raise `TypeError` by default.

        Args:
            value: The value
            value_name: An identifier used for error messages.
                Defaults to the value returned by `cls._get_value_name()`.

        Returns:
            An instance of this class.

        Raises:
            TypeError: If `value` cannot be converted.
        """
        value_name = value_name or cls.value_name()
        if isinstance(value, cls):
            return value
        if value is None:
            return cls._from_none(value_name)
        if isinstance(value, bool):
            return cls._from_bool(value, value_name)
        if isinstance(value, int):
            return cls._from_int(value, value_name)
        if isinstance(value, float):
            return cls._from_float(value, value_name)
        if isinstance(value, str):
            return cls._from_str(value, value_name)
        if isinstance(value, Mapping):
            return cls._from_mapping(value, value_name)
        if isinstance(value, Sequence):
            return cls._from_sequence(value, value_name)
        if isclass(value):
            return cls._from_class(value, value_name)
        return cls._from_other(value, value_name)

    @classmethod
    def _from_none(cls, value_name: str) -> T:
        """Create an instance of this class from a `None` value.
        The default implementation raises a `TypeError`.
        Override to implement a different behaviour.
        """
        raise TypeError(cls._format_type_error(None, value_name))

    @classmethod
    def _from_bool(cls, value: bool, value_name: str) -> T:
        """Create an instance of this class from a bool value.
        The default implementation raises a `TypeError`.
        Override to implement a different behaviour.
        """
        raise TypeError(cls._format_type_error(value, value_name))

    @classmethod
    def _from_int(cls, value: int, value_name: str) -> T:
        """Create an instance of this class from an int value.
        The default implementation raises a `TypeError`.
        Override to implement a different behaviour.
        """
        raise TypeError(cls._format_type_error(value, value_name))

    @classmethod
    def _from_float(cls, value: float, value_name: str) -> T:
        """Create an instance of this class from a float value.
        The default implementation raises a `TypeError`.
        Override to implement a different behaviour.
        """
        raise TypeError(cls._format_type_error(value, value_name))

    @classmethod
    def _from_str(cls, value: str, value_name: str) -> T:
        """Create an instance of this class from a str value.
        The default implementation raises a `TypeError`.
        Override to implement a different behaviour.
        """
        raise TypeError(cls._format_type_error(value, value_name))

    @classmethod
    def _from_class(cls, value: Type, value_name: str) -> T:
        """Create an instance of this class from a type value.
        The default implementation raises a `TypeError`.
        Override to implement a different behaviour.
        """
        raise TypeError(cls._format_type_error(value, value_name))

    @classmethod
    def _from_other(cls, value: Any, value_name: str) -> T:
        """Create an instance of this class from a value of
        an unknown type.
        The default implementation raises a `TypeError`.
        Override to implement a different behaviour.
        """
        raise TypeError(cls._format_type_error(value, value_name))

    @classmethod
    def _from_mapping(cls, value: Mapping, value_name: str) -> T:
        """Create an instance of this class from a mapping value.
        The default implementation raises a `TypeError`.
        Override to implement a different behaviour.
        """
        raise TypeError(cls._format_type_error(value, value_name))

    @classmethod
    def _from_sequence(cls, value: Sequence, value_name: str) -> T:
        """Create an instance of this class from a sequence value.
        The default implementation raises a `TypeError`.
        Override to implement a different behaviour.
        """
        raise TypeError(cls._format_type_error(value, value_name))

    @classmethod
    def _convert_typed_value(
        cls, value: Any, type_annotation: Any, value_name: str
    ) -> Any:
        """To be used by subclasses that wish to convert a value with
        known type for the target value.

        Args:
            value: The value to convert to an instance of the
               type specified by `type_annotation`.
            type_annotation: The annotation representing the target type.
            value_name: An identifier for `value`.

        Returns:
            The converted value.
        """
        type_origin, type_args = cls._process_annotation(type_annotation)

        if value is None:
            # If value is None, ensure value is nullable.
            nullable = (
                type_origin is Any
                or type_origin is NoneType
                or type_origin is Union
                and (Any in type_args or NoneType in type_args)
            )
            if not nullable:
                raise TypeError(cls._format_type_error(value, value_name))
            return None

        if type_origin is Any:
            # We cannot do any further type checking,
            # therefore return the value as-is
            return value

        if type_origin is Literal:
            # Value must be one of literal arguments
            if value not in type_args:
                raise TypeError(format_message_one_of(value_name, value, type_args))
            return value

        if type_origin is Union:
            # For unions try converting the alternatives.
            # Return the first successfully converted value.
            assert len(type_args) > 0
            errors = []
            for type_arg in type_args:
                try:
                    return cls._convert_typed_value(value, type_arg, value_name)
                except (TypeError, ValueError) as e:
                    errors.append((type_arg, e))
            # Note, the error message constructed here is suboptimal.
            # But we sometimes need all details to trace back to the
            # root cause while conversion failed.
            raise TypeError(
                "all type alternatives failed:\n"
                + "\n".join(f"  {formatannotation(a)} --> {e}" for a, e in errors)
            )

        # if origin is a real type and value is of type origin
        if isclass(type_origin):
            # If origin is also a ValueConstructible, we are happy
            if issubclass(type_origin, ValueConstructible):
                return type_origin.from_value(value, value_name=value_name)

            if isinstance(value, type_origin):
                # If value has a compatible type, check first if we
                # can take care of special types, i.e., mappings and sequences.
                if isinstance(value, (bool, int, float, str)):
                    # We take a shortcut here. However, str test
                    # is important, because str is also a sequence!
                    return value

                if issubclass(type_origin, Mapping):
                    key_type, item_type = type_args if type_args else (Any, Any)
                    mapping_value = {}
                    # noinspection PyUnresolvedReferences
                    for k, v in value.items():
                        if not isinstance(k, key_type):
                            raise TypeError(
                                format_message_type_of(
                                    f"keys of {value_name}", k, key_type
                                )
                            )
                        mapping_value[k] = cls._convert_typed_value(
                            v, item_type, f"{value_name}[{k!r}]"
                        )
                    return mapping_value

                if issubclass(type_origin, Sequence):
                    item_type = type_args[0] if type_args else Any
                    # noinspection PyTypeChecker
                    return [
                        cls._convert_typed_value(v, item_type, f"{value_name}[{i}]")
                        for i, v in enumerate(value)
                    ]
                return value

        raise TypeError(
            format_message_type_of(value_name, value, formatannotation(type_annotation))
        )

    @classmethod
    @lru_cache(maxsize=1000)
    def class_parameters(cls) -> Mapping[str, Parameter]:
        """Get the type-resolved parameters of this class' constructor.
        The method returns a cached value for `cls`.

        Can be used by subclasses to process annotations.
        """
        return get_class_parameters(cls, forward_refs=cls.forward_refs())

    @classmethod
    def forward_refs(cls) -> Optional[Mapping[str, type]]:
        """Get an extra namespace to be used for resolving parameter type hints.
        Called from [ValueConstructible._get_class_parameters][].

        Can be overridden to provide a namespace for resolving type
        forward references for your class.
        Defaults to `None`.
        """
        return None

    @classmethod
    def value_name(cls) -> str:
        """Get an identifier for values that can be used to create
        instances of this class.

        Can be overridden to provide a custom, user-friendly value name.
        Defaults to `"value"`.
        """
        return "value"

    @classmethod
    def value_type_name(cls) -> str:
        """Get a descriptive name for the value types that can
        be used to create instances of this class, e.g., `"Rule | str"`.

        Can be overridden to provide a custom, user-friendly type name.
        Defaults to this class' name.
        """
        return cls.__name__

    @classmethod
    def _process_annotation(
        cls, prop_annotation: Any
    ) -> tuple[type | UnionType, tuple[type | UnionType, ...]]:
        type_origin = get_origin(prop_annotation)
        if type_origin is not None:
            type_origin = Union if type_origin is UnionType else type_origin
            type_args = get_args(prop_annotation)
        else:
            type_origin = prop_annotation
            type_args = ()
        if _IS_PYTHON_3_10:  # pragma: no cover
            forward_refs = cls.forward_refs()
            type_origin = cls._resolve_forward_ref(forward_refs, type_origin)
            type_args = tuple(
                cls._resolve_forward_ref(forward_refs, type_arg)
                for type_arg in type_args
            )
        return type_origin, type_args

    @classmethod
    def _resolve_forward_ref(cls, namespace, ref: Any) -> Any:  # pragma: no cover
        if isinstance(ref, str) and namespace:
            return namespace.get(ref, ref)
        else:
            return ref

    @classmethod
    def _format_type_error(cls, value: Any, value_name: str) -> str:
        return format_message_type_of(value_name, value, cls.value_type_name())


class MappingConstructible(Generic[T], ValueConstructible[T]):
    """A mixin that makes your classes constructible from mappings,
    such as a `dict`.

    The default implementation checks if `value` is already an
    instance of this class. If so, it is returned unchanged.

    It then delegates to various `_from_<type>()` methods which
    all raise `TypeError` by default, except for `_from_mapping()`.
    The latter is overridden to deserialize the items of the given
    mapping into values that will be passed to match constructor
    parameters. Type annotations of the parameters will be used
    to perform a type-safe conversion of the mapping values.

    The major use case for this is constructing instances of this
    class from JSON objects.
    """

    @classmethod
    def _from_mapping(cls, mapping: Mapping, value_name: str) -> T:
        """Create an instance of this class from a mapping value."""

        mapping_keys = set(mapping.keys())
        properties = cls.class_parameters()

        args = []
        kwargs = {}
        for prop_name, prop_param in properties.items():
            if prop_name in mapping:
                mapping_keys.remove(prop_name)

                if prop_param.annotation is Parameter.empty:
                    prop_annotation = Any
                else:
                    prop_annotation = prop_param.annotation

                prop_value = cls._convert_property_value(
                    mapping[prop_name],
                    prop_annotation,
                    prop_name,
                    value_name=f"{value_name}.{prop_name}",
                )
                if prop_param.kind == Parameter.POSITIONAL_ONLY:
                    args.append(prop_value)
                else:
                    kwargs[prop_name] = prop_value
            elif (
                prop_param.default is Parameter.empty
            ) or prop_param.kind == Parameter.POSITIONAL_ONLY:
                raise TypeError(
                    f"missing value for required property {value_name}.{prop_name}"
                    f" of type {cls.value_type_name()}"
                )

        if mapping_keys:
            invalid_keys = tuple(
                filter(lambda k: not isinstance(k, str), mapping.keys())
            )
            if invalid_keys:
                invalid_type = type(invalid_keys[0])
                raise TypeError(
                    f"mappings used to instantiate {value_name}"
                    f" of type {cls.__name__}"
                    f" must have keys of type str,"
                    f" but found key of type {invalid_type.__name__}"
                )

            raise TypeError(
                f"{', '.join(sorted(mapping_keys))}"
                f" {'is not a member' if len(mapping_keys) == 1 else 'are not members'}"
                f" of {value_name} of type {cls.__name__}"
            )

        # noinspection PyArgumentList
        return cls(*args, **kwargs)

    # noinspection PyUnusedLocal
    @classmethod
    def _convert_property_value(
        cls, value: Any, prop_annotation: Any, prop_name: str, value_name: str
    ) -> Any:
        """Convert a property value to the given type.

        May be overridden by subclasses that wish to perform a special
        conversion for a specific property.

        Args:
            value: The mapping's item value to convert to an instance of the
                type specified by `prop_annotation`.
            prop_annotation: The property annotation representing the
                target type.
            prop_name: The property name. May be used by overrides.
            value_name: An identifier for `value`.

        Returns:
            The converted property value.
        """
        return cls._convert_typed_value(
            value,
            prop_annotation,
            value_name=value_name,
        )

    @classmethod
    def value_type_name(cls) -> str:
        """Get a descriptive name for the value types that can
        be used to create instances of this class, e.g., `"Rule | str"`.
        Defaults to `f"{cls.__name__} | dict[str, Any]"`.
        """
        return f"{cls.__name__} | dict[str, Any]"


def get_class_parameters(
    cls, forward_refs: Mapping[str, type] | None = None
) -> Mapping[str, Parameter]:
    """Get the type-resolved parameters of this class' constructor.
    The returned value is cached.

    Args:
        cls: The class to inspect.
        forward_refs: Optional extra namespace from which to
            resolve forward references.

    Returns:
        A mapping from parameter names to parameters.
    """
    # Get the signature of the constructor
    sig = signature(cls.__init__)

    # Resolve annotations
    resolved_hints = get_type_hints(cls.__init__, localns=forward_refs)

    # Process the parameters
    resolved_params = {}
    for i, (name, param) in enumerate(sig.parameters.items()):
        if i > 0:  # Skip `self`
            annotation = resolved_hints.get(name, Parameter.empty)
            resolved_params[name] = Parameter(
                name, param.kind, default=param.default, annotation=annotation
            )

    return resolved_params



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import unittest

from xrlint.util.filepattern import FilePattern


class MinimatchTest(unittest.TestCase):
    def test_basics(self):
        matcher = FilePattern("**/*.h5")
        self.assertEqual("**/*.h5", str(matcher))
        self.assertEqual("FilePattern('**/*.h5')", repr(matcher))
        self.assertTrue(matcher == matcher)
        self.assertFalse(matcher == 5)
        self.assertTrue(matcher == FilePattern("**/*.h5"))
        self.assertFalse(matcher == FilePattern("**/*.nc"))
        self.assertFalse(matcher != FilePattern("**/*.h5"))
        self.assertTrue(matcher != FilePattern("**/*.nc"))
        self.assertTrue(hash(matcher) == hash("**/*.h5"))

    def test_no_magic(self):
        matcher = FilePattern("fod")
        self.assertEqual(True, matcher.match("fod"))
        self.assertEqual(True, matcher.match("fod/"))
        self.assertEqual(False, matcher.match("dir1/fod"))
        self.assertEqual(False, matcher.match("dir1/dir2/fod"))

        matcher = FilePattern("dir1/")
        self.assertEqual(True, matcher.match("dir1"))
        self.assertEqual(True, matcher.match("dir1/"))
        self.assertEqual(False, matcher.match("dir1/fod"))
        self.assertEqual(False, matcher.match("dir1/dir2/fod"))

    def test_question_mark_magic(self):
        matcher = FilePattern("dir?/fod")
        self.assertEqual(True, matcher.match("dir1/fod"))
        self.assertEqual(True, matcher.match("dir2/fod"))
        self.assertEqual(False, matcher.match("dir/fod"))

        matcher = FilePattern("fod.???")
        self.assertEqual(True, matcher.match("fod.ext"))
        self.assertEqual(True, matcher.match("fod.tif"))
        self.assertEqual(False, matcher.match("fod.py"))
        self.assertEqual(False, matcher.match("fod.zarr"))

    def test_star_magic(self):
        matcher = FilePattern("*")
        self.assertEqual(True, matcher.match("fod"))
        self.assertEqual(True, matcher.match("fod.ext"))
        self.assertEqual(True, matcher.match("dir1/"))
        self.assertEqual(True, matcher.match("dir1.ext/"))
        self.assertEqual(False, matcher.match("dir1/fod"))
        self.assertEqual(False, matcher.match("dir1/dir2/fod"))

        matcher = FilePattern("*.ext")
        self.assertEqual(True, matcher.match("fod.ext"))
        self.assertEqual(True, matcher.match("fod.ext/"))
        self.assertEqual(False, matcher.match("fod.ext2"))
        self.assertEqual(False, matcher.match("dir1/fod.ext"))

        matcher = FilePattern("fod.*")
        self.assertEqual(True, matcher.match("fod.ext"))
        self.assertEqual(True, matcher.match("fod.ext/"))
        self.assertEqual(True, matcher.match("fod.ext2"))
        self.assertEqual(False, matcher.match("dir1/fod.ext"))

        matcher = FilePattern("dir1/*")
        self.assertEqual(False, matcher.match("dir1"))
        self.assertEqual(False, matcher.match("dir1/"))
        self.assertEqual(True, matcher.match("dir1/fod"))
        self.assertEqual(False, matcher.match("dir1/dir2/fod"))

        matcher = FilePattern("*/fod")
        self.assertEqual(False, matcher.match("fod"))
        self.assertEqual(False, matcher.match("fod/"))
        self.assertEqual(True, matcher.match("dir1/fod"))
        self.assertEqual(False, matcher.match("dir1/dir2/fod"))

        matcher = FilePattern("fod*")
        self.assertEqual(True, matcher.match("fod"))
        self.assertEqual(False, matcher.match("fod/dir2"))

        matcher = FilePattern("dir*/fod")
        self.assertEqual(True, matcher.match("dir/fod"))
        self.assertEqual(True, matcher.match("dir1/fod"))

        matcher = FilePattern("dir1/*/fod")
        self.assertEqual(False, matcher.match("fod"))
        self.assertEqual(False, matcher.match("fod/"))
        self.assertEqual(False, matcher.match("dir1/fod"))
        self.assertEqual(True, matcher.match("dir1/dir2/fod"))

        matcher = FilePattern("*/*/*/*.ext")
        self.assertEqual(False, matcher.match("fod.ext"))
        self.assertEqual(False, matcher.match("dir1/fod.ext"))
        self.assertEqual(False, matcher.match("dir1/dir2/fod.ext"))
        self.assertEqual(True, matcher.match("dir1/dir2/dir3/fod.ext"))
        self.assertEqual(False, matcher.match("dir1/dir2/dir3/dir4/fod.ext"))

    def test_glob_star_magic(self):
        matcher = FilePattern("**")
        self.assertEqual(True, matcher.match("fod"))
        self.assertEqual(True, matcher.match("fod.ext"))
        self.assertEqual(True, matcher.match("dir1/"))
        self.assertEqual(True, matcher.match("dir1.ext/"))
        self.assertEqual(True, matcher.match("dir1/fod"))
        self.assertEqual(True, matcher.match("dir1/dir2/fod"))

        matcher = FilePattern("**.ext")
        self.assertEqual(True, matcher.match("fod.ext"))
        self.assertEqual(True, matcher.match("fod.ext/"))
        self.assertEqual(False, matcher.match("fod.ext2"))
        self.assertEqual(True, matcher.match("dir1/fod.ext"))

        matcher = FilePattern("fod.**")
        self.assertEqual(True, matcher.match("fod.ext"))
        self.assertEqual(True, matcher.match("fod.ext/"))
        self.assertEqual(True, matcher.match("fod.ext2"))
        self.assertEqual(True, matcher.match("fod.ext2/dir2"))
        self.assertEqual(False, matcher.match("dir1/fod.ext"))
        self.assertEqual(False, matcher.match("dir1/fod.ext/dir3"))

        matcher = FilePattern("dir1/**")
        self.assertEqual(False, matcher.match("dir1"))
        self.assertEqual(False, matcher.match("dir1/"))
        self.assertEqual(True, matcher.match("dir1/fod"))
        self.assertEqual(True, matcher.match("dir1/dir2/fod"))

        matcher = FilePattern("**/fod")
        self.assertEqual(True, matcher.match("fod"))
        self.assertEqual(True, matcher.match("fod/"))
        self.assertEqual(True, matcher.match("dir1/fod"))
        self.assertEqual(True, matcher.match("dir1/dir2/fod"))
        self.assertEqual(True, matcher.match("dir1/dir2/dir3/fod"))

        matcher = FilePattern("dir1/**/fod")
        self.assertEqual(False, matcher.match("fod"))
        self.assertEqual(False, matcher.match("fod/"))
        self.assertEqual(True, matcher.match("dir1/fod"))
        self.assertEqual(True, matcher.match("dir1/dir2/fod"))
        self.assertEqual(True, matcher.match("dir1/dir2/dir3/fod"))

    def test_extension_pattern(self):
        matcher = FilePattern("**/*.ext")
        self.assertEqual(True, matcher.match("fod.ext"))
        self.assertEqual(True, matcher.match("dir1/fod.ext"))
        self.assertEqual(True, matcher.match("dir1/dir2/fod.ext"))
        self.assertEqual(True, matcher.match("/fod.ext"))
        self.assertEqual(True, matcher.match("/dir1/dir2/fod.ext"))
        self.assertEqual(False, matcher.match("fod.nc"))
        self.assertEqual(False, matcher.match("dir1/dir2/fod.nc"))

    def test_special_ops(self):
        matcher = FilePattern("")
        self.assertEqual(True, matcher.empty)
        self.assertEqual(True, matcher.match("file"))
        self.assertEqual(True, matcher.match("dir1/"))
        self.assertEqual(True, matcher.match("dir1/fod"))
        self.assertEqual(True, matcher.match("dir1/dir2/fod"))

        matcher = FilePattern("# comment")
        self.assertEqual(True, matcher.comment)
        self.assertEqual(False, matcher.match("fod"))
        self.assertEqual(False, matcher.match("# comment"))

        matcher = FilePattern("!**/.ext")
        self.assertEqual(True, matcher.negate)
        self.assertEqual(False, matcher.match("fod.ext"))
        self.assertEqual(True, matcher.match("fod.ex"))
        self.assertEqual(True, matcher.match("dir1.ex/dir2"))

        matcher = FilePattern("!**/.ext", flip_negate=True)
        self.assertEqual(True, matcher.negate)
        self.assertEqual(True, matcher.match("fod.ext"))
        self.assertEqual(False, matcher.match("fod.ex"))
        self.assertEqual(False, matcher.match("dir1.ex/dir2"))

        matcher = FilePattern("!!**/.ext")
        self.assertEqual(False, matcher.negate)
        self.assertEqual(True, matcher.match("fod.ext"))
        self.assertEqual(False, matcher.match("fod.ex"))
        self.assertEqual(False, matcher.match("dir1.ex/dir2"))

        matcher = FilePattern("dir/")
        self.assertEqual(True, matcher.dir)
        self.assertEqual(True, matcher.match("dir"))
        self.assertEqual(True, matcher.match("dir/"))
        self.assertEqual(False, matcher.match("dir1"))
        self.assertEqual(False, matcher.match("dir/dir2"))



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import AttrNode, AttrsNode, DatasetNode, DataTreeNode, VariableNode
from xrlint.rule import RuleConfig, RuleExit, RuleOp

from ..constants import DATASET_ROOT_NAME, DATATREE_ROOT_NAME
from .rulectx import RuleContextImpl


def apply_rule(
    ctx: RuleContextImpl,
    rule_id: str,
    rule_config: RuleConfig,
):
    """Apply rule given by `rule_id` to dataset given in
    `context` using rule configuration `rule_config`.
    """
    try:
        rule = ctx.config.get_rule(rule_id)
    except ValueError as e:
        ctx.report(f"{e}", fatal=True)
        return

    if rule_config.severity == 0:
        # rule is off
        return

    with ctx.use_state(severity=rule_config.severity):
        # TODO: validate rule_config.args/kwargs against rule.meta.schema
        # noinspection PyArgumentList
        rule_op: RuleOp = rule.op_class(*rule_config.args, **rule_config.kwargs)
        try:
            if ctx.datatree is not None:
                name = (
                    DATATREE_ROOT_NAME
                    if ctx.file_index is None
                    else f"{DATATREE_ROOT_NAME}[{ctx.file_index}]"
                )
                _visit_datatree_node(
                    rule_op,
                    ctx,
                    DataTreeNode(
                        parent=None, path=name, name=name, datatree=ctx.datatree
                    ),
                )
            else:
                name = (
                    DATASET_ROOT_NAME
                    if ctx.file_index is None
                    else f"{DATASET_ROOT_NAME}[{ctx.file_index}]"
                )
                _visit_dataset_node(
                    rule_op,
                    ctx,
                    DatasetNode(parent=None, path=name, name=name, dataset=ctx.dataset),
                )
        except RuleExit:
            # This is ok, the rule requested it.
            pass


def _visit_datatree_node(rule_op: RuleOp, context: RuleContextImpl, node: DataTreeNode):
    with context.use_state(node=node):
        rule_op.validate_datatree(context, node)
        if node.datatree.is_leaf:
            _visit_dataset_node(
                rule_op,
                context,
                DatasetNode(
                    parent=node,
                    path=f"{node.path}/{node.datatree.name}",
                    name=node.datatree.name,
                    dataset=node.datatree.dataset,
                ),
            )
        else:
            for name, datatree in node.datatree.children.items():
                _visit_datatree_node(
                    rule_op,
                    context,
                    DataTreeNode(
                        parent=node,
                        path=f"{node.path}/{name}",
                        name=name,
                        datatree=datatree,
                    ),
                )


def _visit_dataset_node(rule_op: RuleOp, context: RuleContextImpl, node: DatasetNode):
    with context.use_state(dataset=node.dataset, node=node):
        rule_op.validate_dataset(context, node)
        _visit_attrs_node(
            rule_op,
            context,
            AttrsNode(
                parent=node,
                path=f"{node.path}.attrs",
                attrs=node.dataset.attrs,
            ),
        )
        for name, variable in node.dataset.coords.items():
            _visit_variable_node(
                rule_op,
                context,
                VariableNode(
                    parent=node,
                    path=f"{node.path}.coords[{name!r}]",
                    name=name,
                    array=variable,
                ),
            )
        for name, variable in node.dataset.data_vars.items():
            _visit_variable_node(
                rule_op,
                context,
                VariableNode(
                    parent=node,
                    path=f"{node.path}.data_vars[{name!r}]",
                    name=name,
                    array=variable,
                ),
            )


def _visit_variable_node(rule_op: RuleOp, context: RuleContextImpl, node: VariableNode):
    with context.use_state(node=node):
        rule_op.validate_variable(context, node)
        _visit_attrs_node(
            rule_op,
            context,
            AttrsNode(
                parent=node,
                path=f"{node.path}.attrs",
                attrs=node.array.attrs,
            ),
        )


def _visit_attrs_node(rule_op: RuleOp, context: RuleContextImpl, node: AttrsNode):
    with context.use_state(node=node):
        rule_op.validate_attrs(context, node)
        for name, value in node.attrs.items():
            _visit_attr_node(
                rule_op,
                context,
                AttrNode(
                    parent=node,
                    name=name,
                    value=value,
                    path=f"{node.path}[{name!r}]",
                ),
            )


def _visit_attr_node(rule_op: RuleOp, context: RuleContextImpl, node: AttrNode):
    with context.use_state(node=node):
        rule_op.validate_attr(context, node)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.plugin import new_plugin
from xrlint.version import version

plugin = new_plugin(
    name="xcube",
    version=version,
    ref="xrlint.plugins.xcube:export_plugin",
    docs_url="https://xcube.readthedocs.io/en/latest/cubespec.html",
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from collections.abc import MutableMapping
from dataclasses import dataclass
from inspect import getdoc, isclass
from typing import Any, Callable, Type

from xrlint.util.constructible import MappingConstructible
from xrlint.util.importutil import import_value
from xrlint.util.naming import to_kebab_case
from xrlint.util.serializable import JsonSerializable, JsonValue


@dataclass(kw_only=True)
class OperationMeta(MappingConstructible["OpMetadata"], JsonSerializable):
    """Operation metadata."""

    name: str
    """Operation name."""

    version: str = "0.0.0"
    """Operation version."""

    """Operation description. Optional."""
    description: str | None = None

    schema: dict[str, JsonValue] | list[dict[str, JsonValue]] | None = None
    """JSON Schema used to specify and validate the operation's'
    options, if any.

    It can take the following values:

    - Use `None` (the default) to indicate that the operation
      as no options at all.
    - Use a schema to indicate that the operation
      takes keyword arguments only.
      The schema's JSON type must be `"object"`.
    - Use a list of schemas to indicate that the operation
      takes positional-only arguments.
      If given, the number of schemas in the list specifies the
      number of positional arguments that must be provided by users.
    """

    ref: str | None = None
    """Operation reference.
    Specifies the location from where the operation can be
    dynamically imported.
    Must have the form "<module>:<attr>", if given.
    """


class Operation(MappingConstructible["Operation"], JsonSerializable):
    """A mixin class that is used by operation classes.

    An operation class comprises a `meta` property
    that provides the operation metadata. See [OpMetadata][]
    for its interface definition.

    An `op_class` property holds a class that implements the
    operation's logic.

    Rules, processors, and formatters use this mixin.

    Derived classes should provide a constructor that takes at least
    two keyword arguments:

    - `meta` - the metadata object that describes the operation
    - `op_class` - the class that implements the operation

    The `meta` object's class is expected to be constructible
    from keyword arguments with at least a `name: str` argument.
    `meta` objects should also have a writable `ref: str | None`
    property.
    """

    # noinspection PyUnresolvedReferences
    def to_json(self, value_name: str | None = None) -> str:
        if self.meta.ref:
            return self.meta.ref
        return super().to_json(value_name=value_name)

    @classmethod
    def _from_class(cls, value: Type, value_name: str) -> "Operation":
        # noinspection PyTypeChecker
        if issubclass(value, cls.op_base_class()):
            op_class = value
            try:
                # Note, the value.meta attribute is set by
                # the define_op
                #
                # noinspection PyUnresolvedReferences
                meta = op_class.meta
            except AttributeError:
                raise ValueError(
                    f"missing {cls.value_name()} metadata, apply define_{cls.value_name()}()"
                    f" to class {op_class.__name__}"
                )
            # noinspection PyArgumentList
            return cls(meta=meta, op_class=op_class)
        return super()._from_class(value, value_name)

    @classmethod
    def _from_str(cls, value: str, value_name: str) -> "Operation":
        # noinspection PyTypeChecker
        operator, operator_ref = import_value(
            value,
            cls.op_import_attr_name(),
            factory=cls.from_value,
        )
        # noinspection PyUnresolvedReferences
        operator.meta.ref = operator_ref
        return operator

    @classmethod
    def op_import_attr_name(cls) -> str:
        """Get the default name for the attribute that is used to import
        instances of this class from modules.
        """
        return f"export_{cls.value_name()}"

    @classmethod
    def meta_class(cls) -> Type:
        """Get the class of the instances of the `meta` field.
        Defaults to [OperationMeta][xrlint.operation.OperationMeta].
        """
        return OperationMeta

    @classmethod
    def op_base_class(cls) -> Type:
        """Get the base class from which all instances of the `op_class`
        must derive from.
        """
        return type

    @classmethod
    def value_name(cls) -> str:
        """Get a name that describes the operation, e.g.,
        "rule", "processor", "formatter".
        """
        return "operation"

    @classmethod
    def value_type_name(cls) -> str:
        return f"{cls.__name__} | Type[{cls.op_base_class().__name__}] | dict | str"

    @classmethod
    def define_operation(
        cls,
        op_class: Type | None,
        *,
        registry: MutableMapping[str, "Operation"] | None = None,
        meta_kwargs: dict[str, Any] | None = None,
        **kwargs,
    ) -> Callable[[Type], Type] | "Operation":
        """Defines an operation."""
        meta_kwargs = meta_kwargs or {}

        def _define_op(_op_class: Type, decorated=True) -> Type | "Operation":
            cls._assert_op_class_ok(
                f"decorated {cls.value_name()} component", _op_class
            )

            name = meta_kwargs.pop("name", None)
            if not name:
                name = to_kebab_case(_op_class.__name__)
            description = meta_kwargs.pop("description", None)
            if not description:
                description = getdoc(_op_class)
            schema = meta_kwargs.pop("schema", None)
            if schema is None:
                # TODO: if schema not given,
                #   derive it from _op_class' ctor arguments
                # schema = cls._derive_schema(_op_class)
                pass
            # noinspection PyCallingNonCallable
            meta = cls.meta_class()(
                name=name, description=description, schema=schema, **meta_kwargs
            )

            # Register rule metadata in rule operation class
            _op_class.meta = meta

            # noinspection PyArgumentList
            op_instance = cls(meta=meta, op_class=_op_class, **kwargs)
            if registry is not None:
                # Register rule in rule registry
                registry[name] = op_instance
            if decorated:
                return _op_class
            else:
                return op_instance

        if registry is not None and not isinstance(registry, MutableMapping):
            raise TypeError(
                f"registry must be a MutableMapping, but got {type(registry).__name__}"
            )

        if op_class is not None:
            # passing the op_class means an operation instance is expected
            cls._assert_op_class_ok("op_class", op_class)
            return _define_op(op_class, decorated=False)

        # used as decorator, return closure
        return _define_op

    @classmethod
    def _assert_op_class_ok(cls, value_name: str, op_class: Type):
        if not isclass(op_class):
            raise TypeError(
                f"{value_name} must be a class, but got {type(op_class).__name__}"
            )
        # noinspection PyTypeChecker
        if not issubclass(op_class, cls.op_base_class()):
            raise TypeError(
                f"{value_name} must be a subclass of {cls.op_base_class().__name__},"
                f" but got {op_class.__name__}"
            )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.plugin import Plugin
from xrlint.util.importutil import import_submodules


def export_plugin() -> Plugin:
    from .plugin import plugin

    import_submodules("xrlint.plugins.core.rules")

    plugin.define_config(
        "recommended",
        {
            "name": "recommended",
            "rules": {
                "access-latency": "warn",
                "content-desc": "warn",
                "conventions": "warn",
                "coords-for-dims": "error",
                "grid-mappings": "error",
                "lat-coordinate": "error",
                "lon-coordinate": "error",
                "no-empty-attrs": "warn",
                "no-empty-chunks": "off",
                "time-coordinate": "error",
                "var-desc": "warn",
                "var-flags": "error",
                "var-missing-data": "warn",
                "var-units": "warn",
            },
        },
    )

    plugin.define_config(
        "all",
        {
            "name": "all",
            "rules": {rule_id: "error" for rule_id in plugin.rules.keys()},
        },
    )

    return plugin



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import xarray as xr

from xrlint.plugins.core.rules.no_empty_attrs import NoEmptyAttrs
from xrlint.testing import RuleTest, RuleTester

valid_dataset_1 = xr.Dataset(attrs=dict(title="empty"))
valid_dataset_2 = xr.Dataset(
    attrs=dict(title="v-data"),
    coords={"x": xr.DataArray([0, 0.1, 0.2], dims="x", attrs={"units": "s"})},
    data_vars={"v": xr.DataArray([10, 20, 30], dims="x", attrs={"units": "m/s"})},
)
invalid_dataset_1 = valid_dataset_1.copy()
invalid_dataset_2 = valid_dataset_2.copy()
invalid_dataset_3 = valid_dataset_2.copy()
invalid_dataset_1.attrs = {}
invalid_dataset_2.x.attrs = {}
invalid_dataset_3.v.attrs = {}


NoEmptyAttrsTest = RuleTester.define_test(
    "no-empty-attrs",
    NoEmptyAttrs,
    valid=[
        RuleTest(dataset=valid_dataset_1),
        RuleTest(dataset=valid_dataset_2),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_1, expected=1),
        RuleTest(dataset=invalid_dataset_2, expected=1),
        RuleTest(dataset=invalid_dataset_3, expected=1),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

import pytest
import xarray as xr

from xrlint.node import DatasetNode
from xrlint.rule import RuleContext, RuleOp
from xrlint.testing import RuleTest, RuleTester


class ForceTitle(RuleOp):
    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        title = node.dataset.attrs.get("title")
        if not title:
            ctx.report("Datasets must have a title")


VALID_DATASET_1 = xr.Dataset(attrs=dict(title="OC-Climatology"))
VALID_DATASET_2 = xr.Dataset(attrs=dict(title="SST-Climatology"))
INVALID_DATASET_1 = xr.Dataset()
INVALID_DATASET_2 = xr.Dataset(attrs=dict(title=""))


# noinspection PyMethodMayBeStatic
class RuleTesterTest(TestCase):
    def test_ok(self):
        tester = RuleTester(rules={"testing/force-title": "error"})
        tester.run(
            "force-title",
            ForceTitle,
            valid=[
                RuleTest(dataset=VALID_DATASET_1),
                RuleTest(dataset=VALID_DATASET_2),
            ],
            invalid=[
                RuleTest(dataset=INVALID_DATASET_1, expected=1),
                RuleTest(
                    dataset=INVALID_DATASET_2, expected=["Datasets must have a title"]
                ),
            ],
        )

    def test_raises_valid(self):
        tester = RuleTester(rules={"testing/force-title": "error"})
        with pytest.raises(
            AssertionError,
            match=(
                "Rule 'force-title': test_valid_2:"
                " expected no problems, but got one error:\nActual message:\n"
                "  0: Datasets must have a title"
            ),
        ):
            tester.run(
                "force-title",
                ForceTitle,
                valid=[
                    RuleTest(dataset=VALID_DATASET_1),
                    RuleTest(dataset=VALID_DATASET_2),
                    RuleTest(dataset=INVALID_DATASET_1),
                ],
            )

    def test_raises_invalid_with_count(self):
        tester = RuleTester(rules={"testing/force-title": "error"})
        with pytest.raises(
            AssertionError,
            match=(
                "Rule 'force-title': test_invalid_0:"
                " expected one problem, but got no problems."
            ),
        ):
            tester.run(
                "force-title",
                ForceTitle,
                invalid=[
                    RuleTest(dataset=VALID_DATASET_1, expected=1),
                ],
            )

    def test_raises_valid_with_count(self):
        tester = RuleTester(rules={"testing/force-title": "error"})
        with pytest.raises(
            AssertionError,
            match=(
                "Rule 'force-title':"
                " test_invalid_raises_valid_with_count:"
                " expected one problem, but got no problems."
            ),
        ):
            tester.run(
                "force-title",
                ForceTitle,
                invalid=[
                    RuleTest(
                        dataset=VALID_DATASET_1,
                        expected=1,
                        name="raises_valid_with_count",
                    ),
                ],
            )

    def test_raises_invalid_with_matching_message(self):
        tester = RuleTester(rules={"testing/force-title": "error"})
        with pytest.raises(
            AssertionError,
            match=(
                "Rule 'force-title':"
                " test_invalid_raises_invalid_with_matching_message:"
                " expected one problem, but got no problems:\n"
                "Expected message:\n"
                "  0: Datasets must have a title"
            ),
        ):
            tester.run(
                "force-title",
                ForceTitle,
                invalid=[
                    RuleTest(
                        dataset=VALID_DATASET_1,
                        expected=["Datasets must have a title"],
                        name="raises_invalid_with_matching_message",
                    ),
                ],
            )

    def test_raises_invalid_with_mismatching_message(self):
        tester = RuleTester(rules={"testing/force-title": "error"})
        with pytest.raises(
            AssertionError,
            match=(
                "Rule 'force-title':"
                " test_invalid_raises_invalid_with_mismatching_message:"
                " got one error as expected, but encountered message mismatch:\n"
                "Message 0:\n"
                "  Expected: Batasets bust bave a bitle\n"
                "  Actual: Datasets must have a title"
            ),
        ):
            tester.run(
                "force-title",
                ForceTitle,
                invalid=[
                    RuleTest(
                        dataset=INVALID_DATASET_1,
                        expected=["Batasets bust bave a bitle"],
                        name="raises_invalid_with_mismatching_message",
                    ),
                ],
            )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import xarray as xr

from xrlint.plugins.core.rules.conventions import Conventions
from xrlint.testing import RuleTest, RuleTester

valid_dataset_0 = xr.Dataset(attrs=dict(Conventions="CF-1.10"))

invalid_dataset_0 = xr.Dataset()
invalid_dataset_1 = xr.Dataset(attrs=dict(Conventions=1.12))
invalid_dataset_2 = xr.Dataset(attrs=dict(Conventions="CF 1.10"))


ConventionsTest = RuleTester.define_test(
    "conventions",
    Conventions,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_0, kwargs={"match": r"CF-.*"}),
    ],
    invalid=[
        RuleTest(
            dataset=invalid_dataset_0,
            expected=["Missing attribute 'Conventions'."],
        ),
        RuleTest(
            dataset=invalid_dataset_1,
            expected=["Invalid attribute 'Conventions': 1.12."],
        ),
        RuleTest(
            dataset=invalid_dataset_2,
            kwargs={"match": r"CF-.*"},
            expected=[
                "Invalid attribute 'Conventions': 'CF 1.10' doesn't match 'CF-.*'."
            ],
        ),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from typing import Any

import numpy as np

from xrlint.node import VariableNode
from xrlint.plugins.core.plugin import plugin
from xrlint.rule import RuleContext, RuleOp

FLAG_MEANINGS = "flag_meanings"
FLAG_VALUES = "flag_values"
FLAG_MASKS = "flag_masks"


@plugin.define_rule(
    "var-flags",
    version="1.0.0",
    type="suggestion",
    description=(
        "Validate attributes 'flag_values', 'flag_masks' and 'flag_meanings'"
        " that make variables that contain flag values self describing. "
    ),
    docs_url="https://cfconventions.org/cf-conventions/cf-conventions.html#flags",
)
class VarFlags(RuleOp):
    def validate_variable(self, ctx: RuleContext, node: VariableNode):
        flag_values = node.array.attrs.get(FLAG_VALUES)
        flag_masks = node.array.attrs.get(FLAG_MASKS)
        flag_meanings = node.array.attrs.get(FLAG_MEANINGS)

        has_values = flag_values is not None
        has_masks = flag_masks is not None
        has_meanings = flag_meanings is not None

        flag_count: int | None = None

        if has_values:
            flag_count = _validate_flag_values(
                ctx,
                flag_values,
                has_meanings,
            )

        if has_masks:
            flag_count = _validate_flag_masks(
                ctx,
                flag_masks,
                has_meanings,
                flag_count,
            )

        if has_meanings:
            _validate_flag_meanings(
                ctx,
                flag_meanings,
                has_values,
                has_masks,
                flag_count,
            )

        if has_values and has_masks:
            _validate_variable(
                ctx,
                node.array.dtype,
            )


def _validate_flag_values(
    ctx: RuleContext, flag_values: Any, has_meanings: bool
) -> int | None:
    if not has_meanings:
        ctx.report(
            f"Missing attribute {FLAG_MEANINGS!r} to explain attribute {FLAG_VALUES!r}."
        )
    type_ok, flag_count = _check_values(flag_values)
    if not type_ok or flag_count is None:
        ctx.report(
            f"Attribute {FLAG_VALUES!r} must be a"
            " 1-d array of integers with length >= 1."
        )
    return flag_count


def _validate_flag_masks(
    ctx: RuleContext, flag_masks: Any, has_meanings: bool, flag_count: int | None
) -> int | None:
    if not has_meanings:
        ctx.report(
            f"Missing attribute {FLAG_MEANINGS!r} to explain attribute {FLAG_MASKS!r}"
        )
    type_ok, flag_masks_count = _check_values(flag_masks)
    if not type_ok or flag_masks_count is None:
        ctx.report(
            f"Attribute {FLAG_MASKS!r} must be a"
            " 1-d array of integers with length >= 1."
        )
    if flag_count is None:
        flag_count = flag_masks_count
    elif flag_masks_count is not None and flag_masks_count != flag_count:
        ctx.report(
            f"Attribute {FLAG_MASKS!r} must have same length"
            f" as attribute {FLAG_VALUES!r}."
        )
    return flag_count


def _validate_flag_meanings(
    ctx: RuleContext,
    flag_meanings: Any,
    has_values: bool,
    has_masks: bool,
    flag_count: int | None,
):
    if not has_values and not has_masks:
        ctx.report(
            f"Missing attribute {FLAG_VALUES!r} or {FLAG_MASKS!r}"
            f" when attribute {FLAG_MASKS!r} is used."
        )
    type_ok, flag_meanings_count = _check_meanings(flag_meanings)
    if not type_ok or flag_meanings_count is None:
        ctx.report(
            f"Attribute {FLAG_MASKS!r} must be a space-separated string"
            f" with at least two entries."
        )
    if (
        flag_meanings_count is not None
        and flag_count is not None
        and flag_meanings_count != flag_count
    ):
        ctx.report(
            f"Attribute {FLAG_MASKS!r} must have same length"
            f" as attributes {FLAG_VALUES!r} or {FLAG_MEANINGS!r}."
        )


def _validate_variable(ctx: RuleContext, var_dtype: np.dtype):
    if not np.issubdtype(var_dtype, np.integer):
        ctx.report(
            f"Flags variable should have an integer data type, was {var_dtype!r}"
        )


def _check_values(values: Any) -> tuple[bool, int | None]:
    if isinstance(values, (tuple, list)) or (
        isinstance(values, np.ndarray) and values.ndim == 1
    ):
        count = len(values)
        return all(isinstance(v, int) for v in values), count if count >= 1 else None
    return False, None


def _check_meanings(meanings: Any):
    if isinstance(meanings, str):
        meanings_list = [m.strip() for m in meanings.split(" ")]
        count = len(meanings_list)
        return True, count if count >= 1 else None
    return False, None



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

from xrlint.constants import SEVERITY_CODE_TO_NAME, SEVERITY_ENUM, SEVERITY_ENUM_TEXT


class ConstantsTest(TestCase):
    def test_computed_values(self):
        self.assertEqual({0: "off", 1: "warn", 2: "error"}, SEVERITY_CODE_TO_NAME)
        self.assertEqual(
            {"off": 0, "warn": 1, "error": 2, 0: 0, 1: 1, 2: 2},
            SEVERITY_ENUM,
        )
        self.assertEqual("'error', 'warn', 'off', 2, 1, 0", SEVERITY_ENUM_TEXT)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

from xrlint.config import ConfigObject
from xrlint.plugin import new_plugin
from xrlint.result import (
    Message,
    Result,
    ResultStats,
    Suggestion,
    get_rules_meta_for_results,
)
from xrlint.rule import RuleMeta, RuleOp


class ResultTest(TestCase):
    # noinspection PyUnusedLocal
    def test_get_rules_meta_for_results(self):
        plugin = new_plugin(name="test")

        @plugin.define_rule("my-rule-1")
        class MyRule1(RuleOp):
            pass

        @plugin.define_rule("my-rule-2")
        class MyRule2(RuleOp):
            pass

        config_obj = ConfigObject(plugins={"test": plugin})
        rules_meta = get_rules_meta_for_results(
            results=[
                Result(
                    file_path="test.zarr",
                    config_object=config_obj,
                    messages=[Message(message="m 1", rule_id="test/my-rule-1")],
                ),
                Result(
                    file_path="test.zarr",
                    config_object=config_obj,
                    messages=[Message(message="m 2", rule_id="test/my-rule-2")],
                ),
                Result(
                    file_path="test.zarr",
                    config_object=config_obj,
                    messages=[Message(message="m 3", rule_id="test/my-rule-1")],
                ),
                Result(
                    file_path="test.zarr",
                    config_object=config_obj,
                    messages=[Message(message="m 4", rule_id="test/my-rule-2")],
                ),
            ]
        )

        self.assertIsInstance(rules_meta, dict)
        self.assertEqual(2, len(rules_meta))
        for k, v in rules_meta.items():
            self.assertIsInstance(k, str)
            self.assertIsInstance(v, RuleMeta)
        self.assertEqual(
            {"test/my-rule-1", "test/my-rule-2"},
            set(rules_meta.keys()),
        )

    def test_repr_html(self):
        result = Result(
            file_path="test.zarr",
            config_object=ConfigObject(),
            messages=[],
        )
        html = result._repr_html_()
        self.assertIsInstance(html, str)
        self.assertIn("ok", html)
        self.assertIn("<div", html)

        result = Result(
            file_path="test.zarr",
            config_object=ConfigObject(),
            messages=[Message(message="m 1", rule_id="test/my-rule-1")],
        )
        html = result._repr_html_()
        self.assertIsInstance(html, str)
        self.assertIn("<table>", html)
        self.assertIn("</table>", html)


class SuggestionTest(TestCase):
    # noinspection PyUnusedLocal
    def test_from_value(self):
        self.assertEqual(
            Suggestion("Use xr.transpose()"),
            Suggestion.from_value("Use xr.transpose()"),
        )

        suggestion = Suggestion("Use xr.transpose()")
        self.assertIs(suggestion, Suggestion.from_value(suggestion))


class ResultStatsTest(TestCase):
    def test_collect(self):
        stats = ResultStats()

        self.assertEqual(0, stats.error_count)
        self.assertEqual(0, stats.warning_count)
        self.assertEqual(0, stats.result_count)

        results = [
            Result(
                file_path="test.zarr",
                messages=[
                    Message("R1 M1", severity=1),
                    Message("R1 M2", severity=2),
                ],
            ),
            Result(
                file_path="test.zarr",
                messages=[
                    Message("R2 M1", severity=2),
                ],
            ),
            Result(
                file_path="test.zarr",
                messages=[
                    Message("R3 M1", severity=1),
                    Message("R3 M2", severity=2),
                    Message("R3 M3", severity=2),
                ],
            ),
        ]

        results2 = list(stats.collect(results))

        self.assertEqual(results, results2)
        self.assertEqual(4, stats.error_count)
        self.assertEqual(2, stats.warning_count)
        self.assertEqual(3, stats.result_count)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import numpy as np
import xarray as xr

from xrlint.plugins.xcube.rules.time_naming import TimeNaming
from xrlint.testing import RuleTest, RuleTester


def make_dataset(time_var: str, time_dim: str | None = None):
    time_dim = time_dim or time_var
    dims = [time_dim, "y", "x"]
    n = 3
    return xr.Dataset(
        attrs=dict(title="v-data"),
        coords={
            "x": xr.DataArray(np.linspace(0, 1, n), dims="x", attrs={"units": "m"}),
            "y": xr.DataArray(np.linspace(0, 1, n), dims="y", attrs={"units": "m"}),
            time_var: xr.DataArray(
                list(range(n)),
                dims=time_dim,
                attrs={"units": "days since 2010-05-01 UTC", "calendar": "gregorian"},
            ),
        },
        data_vars={
            "chl": xr.DataArray(
                np.random.random((n, n, n)), dims=dims, attrs={"units": "mg/m^-3"}
            ),
            "tsm": xr.DataArray(
                np.random.random((n, n, n)), dims=dims, attrs={"units": "mg/m^-3"}
            ),
        },
    )


valid_dataset_0 = xr.Dataset()
valid_dataset_1 = make_dataset("time")

# Not ok, because time coord not called time
invalid_dataset_0 = make_dataset("t")

# Not ok, because no units
invalid_dataset_1 = make_dataset("time")
del invalid_dataset_1.time.attrs["units"]

# Not ok, because no calendar
invalid_dataset_2 = make_dataset("time")
del invalid_dataset_2.time.attrs["calendar"]

# Not ok, because invalid unit
invalid_dataset_3 = make_dataset("time")
invalid_dataset_3.time.attrs["units"] = "meters"

# Not ok, because coordinate 'time' should have dim 'time'
invalid_dataset_4 = make_dataset("time", "t0")

# Not ok, because coordinate 't0' should be named 'time'
invalid_dataset_5 = make_dataset("t0", "time")

TimeNamingTest = RuleTester.define_test(
    "time-naming",
    TimeNaming,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_1),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_0, expected=1),
        RuleTest(dataset=invalid_dataset_1, expected=1),
        RuleTest(dataset=invalid_dataset_2, expected=1),
        RuleTest(dataset=invalid_dataset_3, expected=1),
        RuleTest(dataset=invalid_dataset_4, expected=1),
        RuleTest(dataset=invalid_dataset_5, expected=2),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import AttrsNode
from xrlint.plugins.core.plugin import plugin
from xrlint.result import Suggestion
from xrlint.rule import RuleContext, RuleOp


@plugin.define_rule(
    "no-empty-attrs",
    version="1.0.0",
    type="suggestion",
    description="Every dataset element should have metadata that describes it.",
)
class NoEmptyAttrs(RuleOp):
    def validate_attrs(self, ctx: RuleContext, node: AttrsNode):
        if not node.attrs:
            ctx.report(
                "Missing metadata, attributes are empty.",
                suggestions=[
                    Suggestion(
                        desc=(
                            "Make sure to add appropriate metadata"
                            " attributes to dataset elements."
                        )
                    )
                ],
            )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import xarray as xr

from xrlint.plugins.core.rules.no_empty_chunks import NoEmptyChunks
from xrlint.testing import RuleTest, RuleTester

# valid, because it is not chunked
valid_dataset_0 = xr.Dataset(attrs=dict(title="OC-Climatology"))
valid_dataset_0.encoding["source"] = "test.zarr"
valid_dataset_0["sst"] = xr.DataArray([273, 274, 272], dims="time")
valid_dataset_0["sst"].encoding["_FillValue"] = 0
valid_dataset_0["sst"].encoding["chunks"] = [3]
# valid, because it does not apply
valid_dataset_1 = valid_dataset_0.copy()
del valid_dataset_1.encoding["source"]
# valid, because it does not apply
valid_dataset_2 = valid_dataset_0.copy()
valid_dataset_2.encoding["source"] = "test.nc"

# valid, because it does not apply
invalid_dataset_0 = valid_dataset_0.copy()
invalid_dataset_0["sst"].encoding["chunks"] = [1]

NoEmptyChunksTest = RuleTester.define_test(
    "no-empty-chunks",
    NoEmptyChunks,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_1),
        RuleTest(dataset=valid_dataset_2),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_0, expected=1),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import json
from typing import Any
from unittest import TestCase

import fsspec
import pytest
import xarray as xr

from tests.plugins.xcube.helpers import make_cube_levels
from xrlint.plugins.xcube.constants import ML_INFO_ATTR
from xrlint.plugins.xcube.processors.mldataset import MultiLevelDatasetProcessor
from xrlint.plugins.xcube.util import LevelInfo, LevelsMeta
from xrlint.result import Message


class MultiLevelDatasetProcessorTest(TestCase):
    levels_name = "xrlint-test"
    levels_dir = f"memory://{levels_name}.levels"

    x_size = nx = 720
    y_size = ny = 360
    time_size = 3
    num_levels = 4

    meta_path = f"{levels_dir}/.zlevels"
    meta_content = {
        "version": "1.0",
        "num_levels": num_levels,
        "use_saved_levels": False,
        # tile_size is optional
        "agg_methods": {
            "chl": "mean",
        },
    }

    @classmethod
    def setUpClass(cls):
        fs, _ = fsspec.core.url_to_fs(cls.levels_dir)
        cls.fs: fsspec.AbstractFileSystem = fs

    def setUp(self):
        self._delete_files()
        self.fs.mkdir(self.levels_dir)

        level_datasets = make_cube_levels(
            self.num_levels, self.x_size, self.y_size, self.time_size
        )
        for level, dataset in enumerate(level_datasets):
            dataset.to_zarr(f"{self.levels_dir}/{level}.zarr")

        with self.fs.open(self.meta_path, mode="wt") as stream:
            json.dump(self.meta_content, stream, indent=2)

    def tearDown(self):
        self._delete_files()

    def _delete_files(self):
        if self.fs.exists(self.levels_dir):
            self.fs.delete(self.levels_dir, recursive=True)

    def assert_levels_ok(
        self, datasets: Any, expect_meta: bool = False, expect_link: bool = False
    ):
        self.assertIsInstance(datasets, list)
        self.assertEqual(self.num_levels, len(datasets))
        for i, (dataset, file_path) in enumerate(datasets):
            if expect_link and i == 0:
                self.assertEqual(f"memory://{self.levels_name}.zarr", file_path)
            else:
                self.assertEqual(
                    f"memory://{self.levels_name}.levels/{i}.zarr", file_path
                )
            level_info = dataset.attrs.get(ML_INFO_ATTR)
            self.assertIsInstance(level_info, LevelInfo)
            self.assertEqual(i, level_info.level)
            self.assertEqual(self.num_levels, level_info.num_levels)
            self.assertIsInstance(level_info.datasets, list)
            self.assertEqual(self.num_levels, len(level_info.datasets))
            meta = level_info.meta
            if expect_meta:
                self.assertIsInstance(meta, LevelsMeta)
                self.assertEqual("1.0", meta.version)
                self.assertEqual(self.num_levels, meta.num_levels)
                self.assertEqual(False, meta.use_saved_levels)
                self.assertEqual(None, meta.tile_size)
                self.assertEqual({"chl": "mean"}, meta.agg_methods)
            else:
                self.assertIsNone(meta)

    def preprocess(self) -> list[tuple[xr.Dataset, str]]:
        processor = MultiLevelDatasetProcessor()
        return processor.preprocess(self.levels_dir, {})

    def test_preprocess(self):
        datasets = self.preprocess()
        self.assert_levels_ok(datasets, expect_meta=True)

    def test_preprocess_no_meta(self):
        self.fs.delete(self.meta_path)
        datasets = self.preprocess()
        self.assert_levels_ok(datasets, expect_meta=False)

    def test_preprocess_with_link(self):
        self.fs.copy(
            f"{self.levels_dir}/0.zarr",
            f"memory://{self.levels_name}.zarr",
            recursive=True,
        )
        self.fs.delete(f"{self.levels_dir}/0.zarr", recursive=True)
        self.fs.write_text(f"{self.levels_dir}/0.link", f"../{self.levels_name}.zarr")
        datasets = self.preprocess()
        self.assert_levels_ok(datasets, expect_meta=True, expect_link=True)

    def test_preprocess_fail_empty(self):
        for i in range(self.num_levels):
            self.fs.delete(f"{self.levels_dir}/{i}.zarr", recursive=True)
        with pytest.raises(ValueError, match="empty multi-level dataset"):
            self.preprocess()

    def test_preprocess_fail_missing(self):
        self.fs.delete(f"{self.levels_dir}/1.zarr", recursive=True)
        with pytest.raises(
            ValueError, match="missing dataset for level 1 in multi-level dataset"
        ):
            self.preprocess()

    def test_postprocess(self):
        processor = MultiLevelDatasetProcessor()
        ml0 = [Message("m00"), Message("m01")]
        ml1 = [Message("10"), Message("m11"), Message("m12")]
        messages = processor.postprocess(
            [
                ml0,
                ml1,
            ],
            self.levels_dir,
        )
        self.assertEqual([*ml0, *ml1], messages)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import re

from xrlint.node import DatasetNode
from xrlint.plugins.core.plugin import plugin
from xrlint.rule import RuleContext, RuleExit, RuleOp
from xrlint.util.schema import schema


@plugin.define_rule(
    "conventions",
    version="1.0.0",
    type="suggestion",
    description=(
        "Datasets should identify the applicable conventions"
        " using the `Conventions` attribute.\n"
        " The rule has an optional configuration parameter `match` which"
        " is a regex pattern that the value of the `Conventions` attribute"
        " must match, if any. If not provided, the rule just verifies"
        " that the attribute exists and whether it is a character string."
    ),
    docs_url=(
        "https://cfconventions.org/cf-conventions/cf-conventions.html"
        "#identification-of-conventions"
    ),
    schema=schema(
        "object",
        properties={
            "match": schema("string", title="Regex pattern"),
        },
    ),
)
class Conventions(RuleOp):
    def __init__(self, match: str | None = None):
        self.match = re.compile(match) if match else None

    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        if "Conventions" not in node.dataset.attrs:
            ctx.report("Missing attribute 'Conventions'.")
        else:
            value = node.dataset.attrs.get("Conventions")
            if not isinstance(value, str) and value:
                ctx.report(f"Invalid attribute 'Conventions': {value!r}.")
            elif self.match is not None and not self.match.match(value):
                ctx.report(
                    f"Invalid attribute 'Conventions':"
                    f" {value!r} doesn't match {self.match.pattern!r}."
                )
        raise RuleExit



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.plugin import Plugin
from xrlint.plugins.xcube.constants import ML_FILE_PATTERN
from xrlint.util.importutil import import_submodules


def export_plugin() -> Plugin:
    from .plugin import plugin

    import_submodules("xrlint.plugins.xcube.rules")
    import_submodules("xrlint.plugins.xcube.processors")

    common_configs = [
        {
            "plugins": {
                "xcube": plugin,
            },
        },
        {
            # Add *.levels to globally included list of file types
            "files": [ML_FILE_PATTERN],
        },
        {
            # Specify a processor for *.levels files
            "files": [ML_FILE_PATTERN],
            "processor": "xcube/multi-level-dataset",
        },
    ]

    plugin.define_config(
        "recommended",
        [
            *common_configs,
            {
                "rules": {
                    "xcube/any-spatial-data-var": "error",
                    "xcube/cube-dims-order": "error",
                    "xcube/data-var-colors": "warn",
                    "xcube/dataset-title": "error",
                    "xcube/grid-mapping-naming": "warn",
                    "xcube/increasing-time": "error",
                    "xcube/lat-lon-naming": "error",
                    "xcube/ml-dataset-meta": "error",
                    "xcube/ml-dataset-time": "warn",
                    "xcube/ml-dataset-xy": "error",
                    "xcube/no-chunked-coords": "warn",
                    "xcube/single-grid-mapping": "error",
                    "xcube/time-naming": "error",
                },
            },
        ],
    )

    plugin.define_config(
        "all",
        [
            *common_configs,
            {
                "rules": {
                    f"xcube/{rule_id}": "error" for rule_id in plugin.rules.keys()
                },
            },
        ],
    )

    return plugin



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from functools import cached_property
from typing import TYPE_CHECKING, Any, TypeAlias, Union

from xrlint.constants import CORE_PLUGIN_NAME
from xrlint.util.constructible import MappingConstructible, ValueConstructible
from xrlint.util.filefilter import FileFilter
from xrlint.util.merge import merge_arrays, merge_dicts, merge_set_lists, merge_values
from xrlint.util.serializable import JsonSerializable

if TYPE_CHECKING:  # pragma: no cover
    # make IDEs and flake8 happy
    from xrlint.plugin import Plugin
    from xrlint.processor import ProcessorOp
    from xrlint.rule import Rule, RuleConfig


ConfigObjectLike: TypeAlias = Union[
    "ConfigObject",
    Mapping[str, Any],
    None,
]
"""Type alias for values that can represent configuration objects.
Can be either a [ConfigObject][xrlint.config.ConfigObject] instance,
or a mapping (e.g. `dict`) with properties defined by `ConfigObject`,
or `None` (empty configuration object).
"""

ConfigLike: TypeAlias = Union[
    "Config",
    ConfigObjectLike,
    str,
    Sequence[ConfigObjectLike | str],
]
"""Type alias for values that can represent configurations.
Can be either a [Config][xrlint.config.Config] instance,
or a [configuration object like][xrlint.config.ConfigObjectLike] value,
or a named plugin configuration, or a sequence of the latter two.
"""


def get_core_plugin() -> "Plugin":
    """Get the fully imported, populated core plugin."""
    from xrlint.plugins.core import export_plugin

    return export_plugin()


def get_core_config_object() -> "ConfigObject":
    """Create a configuration object that includes the core plugin.

    Returns:
        A new `Config` object
    """
    return ConfigObject(plugins={CORE_PLUGIN_NAME: get_core_plugin()})


def split_config_spec(config_spec: str) -> tuple[str, str]:
    """Split a configuration specification into plugin name
    and configuration item name.
    """
    return (
        config_spec.split("/", maxsplit=1)
        if "/" in config_spec
        else (CORE_PLUGIN_NAME, config_spec)
    )


@dataclass(frozen=True, kw_only=True)
class ConfigObject(MappingConstructible, JsonSerializable):
    """Configuration object.

    Configuration objects are the items managed by a
    [configuration][xrlint.config.Config].

    You should not use the class constructor directly.
    Instead, use the `ConfigObject.from_value()` function.
    """

    name: str | None = None
    """A name for the configuration object.
    This is used in error messages and config inspector to help identify
    which configuration object is being used.
    """

    files: list[str] | None = None
    """An array of glob patterns indicating the files that the
    configuration object should apply to. If not specified,
    the configuration object applies to all files matched
    by any other configuration object.

    When a configuration object contains only the files property
    without accompanying rules or settings, it effectively acts as
    a _global file filter_. This means that XRLint will recognize
    and process only the files matching these patterns, thereby
    limiting its scope to the specified files. The inbuilt
    global file filters are `["**/*.zarr", "**/*.nc"]`.
    """

    ignores: list[str] | None = None
    """An array of glob patterns indicating the files that the
    configuration object should not apply to. If not specified,
    the configuration object applies to all files matched by `files`.
    If `ignores` is used without any other keys in the configuration
    object, then the patterns act as _global ignores_.
    """

    linter_options: dict[str, Any] | None = None
    """A dictionary containing options related to the linting process."""

    opener_options: dict[str, Any] | None = None
    """A dictionary containing options that are passed to
    the dataset opener.
    """

    processor: Union["ProcessorOp", str, None] = None
    """Either an object compatible with the `ProcessorOp`
    interface or a string indicating the name of a processor inside
    of a plugin (i.e., `"pluginName/processorName"`).
    """

    plugins: dict[str, "Plugin"] | None = None
    """A dictionary containing a name-value mapping of plugin names to
    plugin objects. When `files` is specified, these plugins are only
    available to the matching files.
    """

    rules: dict[str, "RuleConfig"] | None = None
    """A dictionary containing the configured rules.
    When `files` or `ignores` are specified, these rule configurations
    are only available to the matching files.
    """

    settings: dict[str, Any] | None = None
    """A dictionary containing name-value pairs of information
    that should be available to all rules.
    """

    @cached_property
    def file_filter(self) -> FileFilter:
        """The file filter specified by this configuration. May be empty."""
        return FileFilter.from_patterns(self.files, self.ignores)

    @cached_property
    def empty(self) -> bool:
        """`True` if this configuration object does not configure anything.
        Note, it could still contribute to a global file filter if its
        `files` and `ignores` options are set."""
        return not (
            self.linter_options
            or self.opener_options
            or self.processor
            or self.plugins
            or self.rules
            or self.settings
        )

    def get_plugin(self, plugin_name: str) -> "Plugin":
        """Get the plugin for given plugin name `plugin_name`."""
        plugin = (self.plugins or {}).get(plugin_name)
        if plugin is None:
            raise ValueError(f"unknown plugin {plugin_name!r}")
        return plugin

    def get_rule(self, rule_id: str) -> "Rule":
        """Get the rule for the given rule identifier `rule_id`.

        Args:
            rule_id: The rule identifier including plugin namespace,
                if any. Format `<rule-name>` (builtin rules) or
                `<plugin-name>/<rule-name>`.

        Returns:
            A `Rule` object.

        Raises:
            ValueError: If either the plugin is unknown in this
                configuration or the rule name is unknown.
        """
        plugin_name, rule_name = split_config_spec(rule_id)
        plugin = self.get_plugin(plugin_name)
        rule = (plugin.rules or {}).get(rule_name)
        if rule is None:
            raise ValueError(f"unknown rule {rule_id!r}")
        return rule

    def get_processor_op(
        self, processor_spec: Union["ProcessorOp", str]
    ) -> "ProcessorOp":
        """Get the processor operation for the given
        processor identifier `processor_spec`.
        """
        from xrlint.processor import Processor, ProcessorOp

        if isinstance(processor_spec, ProcessorOp):
            return processor_spec

        plugin_name, processor_name = split_config_spec(processor_spec)
        plugin = self.get_plugin(plugin_name)
        processor: Processor | None = (plugin.processors or {}).get(processor_name)
        if processor is None:
            raise ValueError(f"unknown processor {processor_spec!r}")
        return processor.op_class()

    def merge(self, config: "ConfigObject", name: str = None) -> "ConfigObject":
        return ConfigObject(
            name=name,
            files=self._merge_pattern_lists(self.files, config.files),
            ignores=self._merge_pattern_lists(self.ignores, config.ignores),
            linter_options=self._merge_options(
                self.linter_options, config.linter_options
            ),
            opener_options=self._merge_options(
                self.opener_options, config.opener_options
            ),
            processor=merge_values(self.processor, config.processor),
            plugins=self._merge_plugin_dicts(self.plugins, config.plugins),
            rules=self._merge_rule_dicts(self.rules, config.rules),
            settings=self._merge_options(self.settings, config.settings),
        )

    @classmethod
    def _merge_rule_dicts(
        cls,
        rules1: dict[str, "RuleConfig"] | None,
        rules2: dict[str, "RuleConfig"] | None,
    ) -> dict[str, "RuleConfig"] | None:
        from xrlint.rule import RuleConfig

        def merge_items(r1: RuleConfig, r2: RuleConfig) -> RuleConfig:
            if r1.severity == r2.severity:
                return RuleConfig(
                    r2.severity,
                    merge_arrays(r1.args, r2.args),
                    merge_dicts(r1.kwargs, r2.kwargs),
                )
            return r2

        return merge_dicts(rules1, rules2, merge_items=merge_items)

    @classmethod
    def _merge_pattern_lists(
        cls, patterns1: list[str] | None, patterns2: list[str] | None
    ) -> list[str] | None:
        return merge_set_lists(patterns1, patterns2)

    @classmethod
    def _merge_options(
        cls, settings1: dict[str, Any] | None, settings2: dict[str, Any] | None
    ) -> dict[str, Any] | None:
        return merge_dicts(settings1, settings2, merge_items=merge_values)

    @classmethod
    def _merge_plugin_dicts(
        cls,
        plugins1: dict[str, "Plugin"] | None,
        plugins2: dict[str, "Plugin"] | None,
    ) -> dict[str, "RuleConfig"] | None:
        from xrlint.plugin import Plugin

        def merge_items(_p1: Plugin, p2: Plugin) -> Plugin:
            return p2

        return merge_dicts(plugins1, plugins2, merge_items=merge_items)

    @classmethod
    def _from_none(cls, value_name: str) -> "ConfigObject":
        return ConfigObject()

    @classmethod
    def forward_refs(cls) -> dict[str, type]:
        from xrlint.plugin import Plugin
        from xrlint.processor import Processor, ProcessorOp
        from xrlint.rule import Rule, RuleConfig

        return {
            "Processor": Processor,
            "ProcessorOp": ProcessorOp,
            "Plugin": Plugin,
            "Rule": Rule,
            "RuleConfig": RuleConfig,
        }

    @classmethod
    def value_name(cls) -> str:
        return "config_obj"

    @classmethod
    def value_type_name(cls) -> str:
        return "ConfigObject | dict | None"


@dataclass(frozen=True)
class Config(ValueConstructible, JsonSerializable):
    """Represents a XRLint configuration.
    A `Config` instance basically manages a sequence of
    [configuration objects][xrlint.config.ConfigObject].

    You should not use the class constructor directly.
    Instead, use the `Config.from_value()` function.
    """

    objects: list[ConfigObject] = field(default_factory=list)
    """The configuration objects."""

    def split_global_filter(
        self, default: FileFilter | None = None
    ) -> tuple["Config", FileFilter]:
        """Get a global file filter for this configuration list."""
        global_filter = FileFilter(
            default.files if default else (),
            default.ignores if default else (),
        )
        objects = []
        for co in self.objects:
            if co.empty and not co.file_filter.empty:
                global_filter = global_filter.merge(co.file_filter)
            else:
                objects.append(co)
        return Config(objects=objects), global_filter

    def compute_config_object(self, file_path: str) -> ConfigObject | None:
        """Compute the configuration object for the given file path.

        Args:
            file_path: A dataset file path.

        Returns:
            A `Config` object which may be empty, or `None`
                if `file_path` is not included by any `files` pattern
                or intentionally ignored by global `ignores`.
        """

        config_obj = None
        for co in self.objects:
            if co.file_filter.empty or co.file_filter.accept(file_path):
                config_obj = config_obj.merge(co) if config_obj is not None else co

        if config_obj is None:
            return None

        # Note, computed configurations do not have "files" and "ignores"
        return ConfigObject(
            linter_options=config_obj.linter_options,
            opener_options=config_obj.opener_options,
            processor=config_obj.processor,
            plugins=config_obj.plugins,
            rules=config_obj.rules,
            settings=config_obj.settings,
        )

    @classmethod
    def from_config(
        cls,
        *configs: ConfigLike,
        value_name: str | None = None,
    ) -> "Config":
        """Convert variable arguments of configuration-like objects
        into a new `Config` instance.

        Args:
            *configs: Variable number of configuration-like arguments.
                For more information see the
                [ConfigLike][xrlint.config.ConfigLike] type alias.
            value_name: The value's name used for reporting errors.

        Returns:
            A new `Config` instance.
        """
        value_name = value_name or cls.value_name()
        objects: list[ConfigObject] = []
        plugins: dict[str, Plugin] = {}
        for i, config_like in enumerate(configs):
            new_objects = None
            if isinstance(config_like, str):
                if CORE_PLUGIN_NAME not in plugins:
                    plugins.update({CORE_PLUGIN_NAME: get_core_plugin()})
                new_objects = cls._get_named_config(config_like, plugins).objects
            elif isinstance(config_like, Config):
                new_objects = config_like.objects
            elif isinstance(config_like, (list, tuple)):
                new_objects = cls.from_config(
                    *config_like, value_name=f"{value_name}[{i}]"
                ).objects
            elif config_like:
                new_objects = [
                    ConfigObject.from_value(
                        config_like, value_name=f"{value_name}[{i}]"
                    )
                ]
            if new_objects:
                for co in new_objects:
                    objects.append(co)
                    plugins.update(co.plugins if co.plugins else {})
        return cls(objects=objects)

    @classmethod
    def from_value(cls, value: ConfigLike, value_name: str | None = None) -> "Config":
        """Convert given `value` into a `Config` object.

        If `value` is already a `Config` then it is returned as-is.

        Args:
            value: A configuration-like value. For more information
                see the [ConfigLike][xrlint.config.ConfigLike] type alias.
            value_name: The value's name used for reporting errors.
        Returns:
            A `Config` object.
        """
        if isinstance(value, (ConfigObject, dict)):
            return Config(objects=[ConfigObject.from_value(value)])
        return super().from_value(value, value_name=value_name)

    @classmethod
    def _from_sequence(cls, value: Sequence, value_name: str) -> "Config":
        return cls.from_config(*value, value_name=value_name)

    @classmethod
    def value_name(cls) -> str:
        return "config"

    @classmethod
    def value_type_name(cls) -> str:
        return "Config | ConfigObjectLike | str | Sequence[ConfigObjectLike | str]"

    @classmethod
    def _get_named_config(
        cls, config_spec: str, plugins: dict[str, "Plugin"]
    ) -> "Config":
        plugin_name, config_name = (
            config_spec.split("/", maxsplit=1)
            if "/" in config_spec
            else (CORE_PLUGIN_NAME, config_spec)
        )
        plugin: Plugin | None = plugins.get(plugin_name)
        if plugin is None or not plugin.configs or config_name not in plugin.configs:
            raise ValueError(f"configuration {config_spec!r} not found")
        return Config.from_value(plugin.configs[config_name])



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.plugins.xcube.rules.any_spatial_data_var import AnySpatialDataVar
from xrlint.testing import RuleTest, RuleTester

from .test_grid_mapping_naming import make_dataset

valid_dataset = make_dataset()
invalid_dataset = valid_dataset.drop_vars(["chl", "tsm"])


AnySpatialDataVarTest = RuleTester.define_test(
    "any-spatial-data-var",
    AnySpatialDataVar,
    valid=[
        RuleTest(dataset=valid_dataset),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset, expected=1),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import DatasetNode
from xrlint.plugins.xcube.plugin import plugin
from xrlint.rule import RuleContext, RuleOp


@plugin.define_rule(
    "dataset-title",
    version="1.0.0",
    type="problem",
    description="Datasets should be given a non-empty title.",
    docs_url="https://xcube.readthedocs.io/en/latest/cubespec.html#metadata",
)
class DatasetTitle(RuleOp):
    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        attrs = node.dataset.attrs
        if "title" not in attrs:
            ctx.report("Missing attribute 'title'.")
        elif not attrs["title"]:
            ctx.report(f"Invalid attribute 'title': {attrs['title']!r}")



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import unittest
from pathlib import Path
from typing import Any
from unittest import TestCase

import pytest

from xrlint.cli.config import ConfigError, read_config
from xrlint.config import Config, ConfigObject
from xrlint.rule import RuleConfig

from .helpers import text_file

yaml_text = """
- name: yaml-test
  rules:
    rule-1: 2
    rule-2: "warn"
    rule-3: ["error", {max_size: 500}]
"""


json_text = """
[
    {
        "name": "json-test",
        "rules": {
          "rule-1": 2,
          "rule-2": "warn",
          "rule-3": ["error", {"max_size": 500}]
        }
    }
]
"""

py_text = """
def export_config():
    return [
        {
            "name": "py-test",
            "rules": {
                "rule-1": 2,
                "rule-2": 1,
                "rule-3": [2, {"max_size": 500}]
            }
        }
    ]
"""


# noinspection PyMethodMayBeStatic
class CliConfigTest(TestCase):
    module_no = 1000

    def new_config_py(self):
        CliConfigTest.module_no += 1
        return f"config_{CliConfigTest.module_no}.py"

    def test_read_config_yaml(self):
        with text_file("config.yaml", yaml_text) as config_path:
            config = read_config(config_path)
            self.assert_config_ok(config, "yaml-test")

    def test_read_config_json(self):
        with text_file("config.json", json_text) as config_path:
            config = read_config(config_path)
            self.assert_config_ok(config, "json-test")

    def test_read_config_py(self):
        with text_file(self.new_config_py(), py_text) as config_path:
            config = read_config(config_path)
            self.assert_config_ok(config, "py-test")

    def assert_config_ok(self, config: Any, name: str):
        self.assertEqual(
            Config(
                [
                    ConfigObject(
                        name=name,
                        rules={
                            "rule-1": RuleConfig(2),
                            "rule-2": RuleConfig(1),
                            "rule-3": RuleConfig(2, kwargs={"max_size": 500}),
                        },
                    )
                ]
            ),
            config,
        )

    def test_read_config_invalid_arg(self):
        with pytest.raises(
            TypeError,
            match="configuration file must be of type str|Path|PathLike, but got None",
        ):
            # noinspection PyTypeChecker
            read_config(None)

    def test_read_config_json_with_format_error(self):
        with text_file("config.json", "{") as config_path:
            with pytest.raises(
                ConfigError,
                match=(
                    "config.json:"
                    " Expecting property name enclosed in double quotes:"
                    " line 1 column 2 \\(char 1\\)"
                ),
            ):
                read_config(config_path)

    def test_read_config_yaml_with_format_error(self):
        with text_file("config.yaml", "}") as config_path:
            with pytest.raises(
                ConfigError,
                match="config.yaml: while parsing a block node",
            ):
                read_config(config_path)

    def test_read_config_yaml_with_type_error(self):
        with text_file("config.yaml", "97") as config_path:
            with pytest.raises(
                ConfigError,
                match=(
                    r"config\.yaml\: config must be of type"
                    r" Config \| ConfigObjectLike \| str \| Sequence\[ConfigObjectLike \| str\],"
                    r" but got int"
                ),
            ):
                read_config(config_path)

    def test_read_config_with_unknown_format(self):
        with pytest.raises(
            ConfigError,
            match="config.toml: unsupported configuration file format",
        ):
            read_config("config.toml")

    def test_read_config_py_no_export(self):
        py_code = "x = 42\n"
        with text_file(self.new_config_py(), py_code) as config_path:
            with pytest.raises(
                ConfigError,
                match=(
                    "config_1002.py: attribute 'export_config'"
                    " not found in module 'config_1002'"
                ),
            ):
                read_config(config_path)

    def test_read_config_py_with_value_error(self):
        py_code = "def export_config():\n    raise ValueError('value is useless!')\n"
        with text_file(self.new_config_py(), py_code) as config_path:
            with pytest.raises(
                ValueError,
                match="value is useless!",
            ):
                read_config(config_path)

    def test_read_config_py_with_os_error(self):
        py_code = "def export_config():\n    raise OSError('where is my hat?')\n"
        with text_file(self.new_config_py(), py_code) as config_path:
            with pytest.raises(
                ConfigError,
                match="where is my hat?",
            ):
                read_config(config_path)

    def test_read_config_py_with_invalid_config_list(self):
        py_code = "def export_config():\n    return 42\n"
        with text_file(self.new_config_py(), py_code) as config_path:
            with pytest.raises(
                ConfigError,
                match=(
                    r"\.py: failed converting value of 'config_1003:export_config':"
                    r" config must be of type"
                    r" Config \| ConfigObjectLike \| str \| Sequence\[ConfigObjectLike \| str\],"
                    r" but got int"
                ),
            ):
                read_config(config_path)


class CliConfigResolveTest(unittest.TestCase):
    def test_read_config_py(self):
        self.assert_ok(
            read_config(Path(__file__).parent / "configs" / "recommended.py")
        )

    def test_read_config_json(self):
        self.assert_ok(
            read_config(Path(__file__).parent / "configs" / "recommended.json")
        )

    def test_read_config_yaml(self):
        self.assert_ok(
            read_config(Path(__file__).parent / "configs" / "recommended.yaml")
        )

    def assert_ok(self, config: Config):
        self.assertIsInstance(config, Config)
        self.assertEqual(7, len(config.objects))
        config_obj = config.compute_config_object("test.zarr")
        self.assertIsInstance(config_obj, ConfigObject)
        self.assertEqual(None, config_obj.name)
        self.assertIsInstance(config_obj.plugins, dict)
        self.assertEqual({"xcube"}, set(config_obj.plugins.keys()))
        self.assertIsInstance(config_obj.rules, dict)
        self.assertIn("coords-for-dims", config_obj.rules)
        self.assertIn("xcube/cube-dims-order", config_obj.rules)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import DatasetNode
from xrlint.plugins.xcube.constants import GM_NAMES_TEXT, LAT_NAME, LON_NAME
from xrlint.plugins.xcube.plugin import plugin
from xrlint.plugins.xcube.util import is_spatial_var
from xrlint.rule import RuleContext, RuleOp


@plugin.define_rule(
    "single-grid-mapping",
    version="1.0.0",
    type="problem",
    description=(
        "A single grid mapping shall be used for all"
        " spatial data variables of a datacube."
    ),
    docs_url="https://xcube.readthedocs.io/en/latest/cubespec.html#spatial-reference",
)
class SingleGridMapping(RuleOp):
    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        dataset = node.dataset

        if not dataset.data_vars:
            # Rule applies to dataset with data variables only
            return

        # Get the mapping of grid mapping names to grid-mapped variables
        grid_mapped_vars = {
            str(v.attrs.get("grid_mapping")): str(k)
            for k, v in dataset.data_vars.items()
            if (is_spatial_var(v)) and "grid_mapping" in v.attrs
        }

        # datacubes with geographic CRS do not need an explicit grid mapping
        geo_crs = LON_NAME in dataset.coords and LAT_NAME in dataset.coords
        if geo_crs and not grid_mapped_vars:
            return

        # if there is not a single grid mapping then report it
        if len(grid_mapped_vars) > 1:
            gm_names = "".join(
                [
                    f"var {var_name!r} -> {gm_name!r}"
                    for gm_name, var_name in grid_mapped_vars.items()
                ]
            )
            ctx.report(
                f"Spatial variables refer to multiple grid mappings: {gm_names}.",
                suggestions=[
                    (
                        "Split datacube into multiple datacubes"
                        " each with a single grid mapping."
                    ),
                ],
            )
        elif len(grid_mapped_vars) == 0:
            ctx.report(
                "None of the spatial variables provides a grid mapping.",
                suggestions=[
                    (
                        f"Add a grid mapping coordinate variable named"
                        f" {GM_NAMES_TEXT} to the dataset."
                    ),
                    (
                        "Set attribute 'grid_mapping' of spatial data variables"
                        " to the name of the grid mapping coordinate variable."
                    ),
                ],
            )

        # Note the validity of grid mappings should be covered by
        # core rule "grid-mappings".



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

from xrlint.plugins.xcube.util import is_absolute_path
from xrlint.plugins.xcube.util import resolve_path


class UtilTest(TestCase):
    def test_is_absolute_path(self):
        self.assertTrue(is_absolute_path("/home/forman"))
        self.assertTrue(is_absolute_path("//bcserver2/fs1"))
        self.assertTrue(is_absolute_path("file://home/forman"))
        self.assertTrue(is_absolute_path("s3://xcube-data"))
        self.assertTrue(is_absolute_path(r"C:\Users\Norman"))
        self.assertTrue(is_absolute_path(r"C:/Users/Norman"))
        self.assertTrue(is_absolute_path(r"C:/Users/Norman"))
        self.assertTrue(is_absolute_path(r"\\bcserver2\fs1"))

        self.assertFalse(is_absolute_path(r"data"))
        self.assertFalse(is_absolute_path(r"./data"))
        self.assertFalse(is_absolute_path(r"../data"))

    def test_resolve_path(self):
        self.assertEqual(
            "/home/forman/data", resolve_path("data", root_path="/home/forman")
        )
        self.assertEqual(
            "/home/forman/data", resolve_path("./data", root_path="/home/forman")
        )
        self.assertEqual(
            "/home/data", resolve_path("../data", root_path="/home/forman")
        )
        self.assertEqual("s3://opensr/test.zarr", resolve_path("s3://opensr/test.zarr"))



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.node import DatasetNode
from xrlint.plugins.xcube.plugin import plugin
from xrlint.plugins.xcube.util import is_spatial_var
from xrlint.rule import RuleContext, RuleOp


@plugin.define_rule(
    "any-spatial-data-var",
    version="1.0.0",
    type="problem",
    description="A datacube should have spatial data variables.",
    docs_url=(
        "https://xcube.readthedocs.io/en/latest/cubespec.html#data-model-and-format"
    ),
)
class AnySpatialDataVar(RuleOp):
    def validate_dataset(self, ctx: RuleContext, node: DatasetNode):
        if not any(map(is_spatial_var, node.dataset.data_vars.values())):
            ctx.report("No spatial data variables found.")



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from xrlint.cli.engine import XRLint
from xrlint.config import Config, ConfigLike, ConfigObject, ConfigObjectLike
from xrlint.formatter import (
    Formatter,
    FormatterContext,
    FormatterMeta,
    FormatterOp,
    FormatterRegistry,
)
from xrlint.linter import Linter, new_linter
from xrlint.node import AttrNode, AttrsNode, DatasetNode, Node, VariableNode
from xrlint.plugin import Plugin, PluginMeta, new_plugin
from xrlint.processor import Processor, ProcessorMeta, ProcessorOp, define_processor
from xrlint.result import (
    EditInfo,
    Message,
    Result,
    Suggestion,
    get_rules_meta_for_results,
)
from xrlint.rule import (
    Rule,
    RuleConfig,
    RuleContext,
    RuleExit,
    RuleMeta,
    RuleOp,
    define_rule,
)
from xrlint.testing import RuleTest, RuleTester
from xrlint.version import version

__all__ = [
    "XRLint",
    "Config",
    "ConfigLike",
    "ConfigObject",
    "ConfigObjectLike",
    "Linter",
    "new_linter",
    "EditInfo",
    "Message",
    "Result",
    "Suggestion",
    "get_rules_meta_for_results",
    "Formatter",
    "FormatterContext",
    "FormatterMeta",
    "FormatterOp",
    "FormatterRegistry",
    "AttrNode",
    "AttrsNode",
    "VariableNode",
    "DatasetNode",
    "Node",
    "Plugin",
    "PluginMeta",
    "new_plugin",
    "Processor",
    "ProcessorMeta",
    "ProcessorOp",
    "define_processor",
    "Rule",
    "RuleConfig",
    "RuleContext",
    "RuleExit",
    "RuleMeta",
    "RuleOp",
    "define_rule",
    "RuleTest",
    "RuleTester",
    "version",
]



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from dataclasses import fields, is_dataclass
from typing import Any, Final, Mapping, Sequence, TypeAlias

from xrlint.util.formatting import format_message_type_of

JSON_VALUE_TYPE_NAME: Final = "None | bool | int | float | str | dict | list"

JsonValue: TypeAlias = (
    None | bool | int | float | str | dict[str, "JsonValue"] | list["JsonValue"]
)


class JsonSerializable:
    """A mixin that makes your classes serializable to JSON values
    and JSON-serializable dictionaries.

    It adds two methods:

    * [to_json][JsonSerializable.to_json] converts to JSON values
    * [to_dict][JsonSerializable.to_dict] converts to JSON-serializable
        dictionaries

    """

    def to_json(self, value_name: str | None = None) -> JsonValue:
        """Convert this object into a JSON value.

        The default implementation calls `self.to_dict()` and returns
        its value as-is.
        """
        return self.to_dict(value_name=value_name)

    def to_dict(self, value_name: str | None = None) -> dict[str, JsonValue]:
        """Convert this object into a JSON-serializable dictionary.

        The default implementation naively serializes the non-protected
        attributes of this object's dictionary given by `vars(self)`.
        """
        return self._object_to_json(self, value_name or type(self).__name__)

    @classmethod
    def _value_to_json(cls, value: Any, value_name: str) -> JsonValue:
        if value is None:
            return None
        if isinstance(value, JsonSerializable):
            return value.to_json(value_name=value_name)
        if isinstance(value, bool):
            return bool(value)
        if isinstance(value, int):
            return int(value)
        if isinstance(value, float):
            return float(value)
        if isinstance(value, str):
            return str(value)
        if isinstance(value, Mapping):
            return cls._mapping_to_json(value, value_name)
        if isinstance(value, Sequence):
            return cls._sequence_to_json(value, value_name)
        if isinstance(value, type):
            return repr(value)
        raise TypeError(format_message_type_of(value_name, value, JSON_VALUE_TYPE_NAME))

    @classmethod
    def _object_to_json(cls, value: Any, value_name: str) -> dict[str, JsonValue]:
        if is_dataclass(value):
            _d = {f.name: (f, getattr(value, f.name)) for f in fields(value)}
            d = {k: v for k, (f, v) in _d.items() if v != f.default}
        else:
            d = {
                k: v
                for k, v in vars(value).items()
                if cls._is_non_protected_property_name(k)
            }
        return {k: cls._value_to_json(v, f"{value_name}.{k}") for k, v in d.items()}

    @classmethod
    def _mapping_to_json(
        cls, mapping: Mapping, value_name: str
    ) -> dict[str, JsonValue]:
        return {
            str(k): cls._value_to_json(v, f"{value_name}[{k!r}]")
            for k, v in mapping.items()
        }

    @classmethod
    def _sequence_to_json(cls, sequence: Sequence, value_name: str) -> list[JsonValue]:
        return [
            cls._value_to_json(v, f"{value_name}[{i}]") for i, v in enumerate(sequence)
        ]

    @classmethod
    def _is_non_protected_property_name(cls, key: Any) -> bool:
        return (
            isinstance(key, str)
            and key.isidentifier()
            and not key[0].isupper()
            and not key[0] == "_"
        )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import sys
from os import PathLike
from pathlib import Path
from typing import Any

import fsspec

from xrlint.config import Config
from xrlint.util.formatting import format_message_type_of
from xrlint.util.importutil import ValueImportError, import_value


def read_config(config_path: str | Path | PathLike[str]) -> Config:
    """Read configuration from configuration file.

    Args:
        config_path: configuration file path.

    Returns:
        A `Config` instance.

    Raises:
        TypeError: if `config_path` is not a path-like object
        FileNotFoundError: if configuration file could not be found
        ConfigError: if the configuration could not be read or is
            otherwise invalid.
    """
    if not isinstance(config_path, (str, Path, PathLike)):
        raise TypeError(
            format_message_type_of("config_path", config_path, "str|Path|PathLike")
        )

    try:
        config_like = _read_config_like(str(config_path))
    except FileNotFoundError:
        raise
    except OSError as e:
        raise ConfigError(config_path, e) from e

    try:
        return Config.from_value(config_like)
    except (ValueError, TypeError) as e:
        raise ConfigError(config_path, e) from e


def _read_config_like(config_path: str) -> Any:
    if config_path.endswith(".yml") or config_path.endswith(".yaml"):
        return _read_config_yaml(config_path)
    if config_path.endswith(".json"):
        return _read_config_json(config_path)
    if config_path.endswith(".py"):
        return _read_config_python(config_path)
    raise ConfigError(config_path, "unsupported configuration file format")


def _read_config_yaml(config_path) -> Any:
    import yaml

    with fsspec.open(config_path, mode="r") as f:
        try:
            return yaml.load(f, Loader=yaml.SafeLoader)
        except yaml.YAMLError as e:
            raise ConfigError(config_path, e) from e


def _read_config_json(config_path) -> Any:
    import json

    with fsspec.open(config_path, mode="r") as f:
        try:
            return json.load(f)
        except json.JSONDecodeError as e:
            raise ConfigError(config_path, e) from e


def _read_config_python(config_path: str) -> Any:
    module_path = Path(config_path)

    if not module_path.exists():
        raise FileNotFoundError(f"file not found: {config_path}")

    module_parent = module_path.parent
    module_name = module_path.stem

    old_sys_path = sys.path
    sys.path = [module_parent.as_posix()] + sys.path
    try:
        return import_value(
            module_name,
            "export_config",
            factory=Config.from_value,
        )[0]
    except ValueImportError as e:
        raise ConfigError(config_path, e) from e
    finally:
        sys.path = old_sys_path


class ConfigError(Exception):
    """An error raised if loading of configuration fails."""

    def __init__(self, config_path: str, e: Exception | str | None = None):
        super().__init__(config_path if e is None else f"{config_path}: {e}")



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from typing import Any, Callable


def merge_values(
    value1: Any,
    value2: Any,
    merge_items: Callable[[Any, Any], Any] | None = None,
) -> Any:
    if isinstance(value1, (list, tuple)) and isinstance(value2, (list, tuple)):
        return merge_arrays(value1, value2, merge_items=merge_items)
    if isinstance(value1, dict) and isinstance(value2, dict):
        return merge_dicts(value1, value2, merge_items=merge_items)
    return value2 if value2 is not None else value1


def merge_dicts(
    dct1: dict[str, Any] | None,
    dct2: dict[str, Any] | None,
    merge_items: Callable[[Any, Any], Any] | None = None,
) -> dict[str, Any] | None:
    if dct1 is None:
        return dct2
    if dct2 is None:
        return dct1
    result = {}
    for k1, v1 in dct1.items():
        if k1 in dct2:
            v2 = dct2[k1]
            result[k1] = merge_items(v1, v2) if merge_items is not None else v2
        else:
            result[k1] = v1
    for k2, v2 in dct2.items():
        if k2 not in dct1:
            result[k2] = v2
    return result


def merge_arrays(
    arr1: list | tuple | None,
    arr2: list | tuple | None,
    merge_items: Callable[[Any, Any], Any] | None = None,
) -> list | tuple | None:
    if arr1 is None:
        return arr2
    if arr2 is None:
        return arr1
    n1 = len(arr1)
    n2 = len(arr2)
    result = list(arr1)
    for i, v1 in enumerate(arr1):
        if i < n2:
            v2 = arr2[i]
            result[i] = merge_items(v1, v2) if merge_items is not None else v2
    if n1 < n2:
        result.extend(arr2[n1:])
    return result


def merge_set_lists(set1: list | None, set2: list | None) -> list | None:
    if set1 is None:
        return set2
    if set2 is None:
        return set1
    result = []
    for v in set1 + set2:
        if v not in result:
            result.append(v)
    return result



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

# No other imports here!
from .version import version

__version__ = version



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import numpy as np
import xarray as xr

from xrlint.plugins.xcube.rules.grid_mapping_naming import GridMappingNaming
from xrlint.testing import RuleTest, RuleTester


def make_dataset():
    return xr.Dataset(
        attrs=dict(title="OC Data"),
        coords={
            "x": xr.DataArray(np.linspace(0, 1, 4), dims="x", attrs={"units": "m"}),
            "y": xr.DataArray(np.linspace(0, 1, 3), dims="y", attrs={"units": "m"}),
            "time": xr.DataArray([2022, 2021], dims="time", attrs={"units": "years"}),
            "crs": xr.DataArray(
                0,
                attrs={
                    "grid_mapping_name": "latitude_longitude",
                    "semi_major_axis": 6371000.0,
                    "inverse_flattening": 0,
                },
            ),
        },
        data_vars={
            "chl": xr.DataArray(
                np.random.random((2, 3, 4)),
                dims=["time", "y", "x"],
                attrs={"units": "mg/m^-3", "grid_mapping": "crs"},
            ),
            "tsm": xr.DataArray(
                np.random.random((2, 3, 4)),
                dims=["time", "y", "x"],
                attrs={"units": "mg/m^-3", "grid_mapping": "crs"},
            ),
        },
    )


valid_dataset_1 = make_dataset()
valid_dataset_2 = valid_dataset_1.rename({"crs": "spatial_ref"})
valid_dataset_3 = valid_dataset_1.drop_vars("crs")

invalid_dataset_1 = valid_dataset_1.rename({"crs": "gm"})


GridMappingNamingTest = RuleTester.define_test(
    "grid-mapping-naming",
    GridMappingNaming,
    valid=[
        RuleTest(dataset=valid_dataset_1),
        RuleTest(dataset=valid_dataset_2),
        RuleTest(dataset=valid_dataset_3),
    ],
    invalid=[
        RuleTest(dataset=invalid_dataset_1, expected=1),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from abc import ABC, abstractmethod
from collections.abc import MutableMapping, Sequence
from dataclasses import dataclass, field
from typing import Any, Callable, Literal, Type

import xarray as xr

from xrlint.constants import SEVERITY_ENUM, SEVERITY_ENUM_TEXT
from xrlint.node import AttrNode, AttrsNode, DatasetNode, DataTreeNode, VariableNode
from xrlint.operation import Operation, OperationMeta
from xrlint.result import Suggestion
from xrlint.util.constructible import ValueConstructible
from xrlint.util.formatting import format_message_one_of
from xrlint.util.serializable import JsonSerializable


class RuleContext(ABC):
    """The context passed to a [RuleOp][xrlint.rule.RuleOp] instance.

    Instances of this interface are passed to the validation
    methods of your `RuleOp`.
    There should be no reason to create instances of this class
    yourself.
    """

    @property
    @abstractmethod
    def file_path(self) -> str:
        """The current dataset's file path."""

    @property
    @abstractmethod
    def settings(self) -> dict[str, Any]:
        """Applicable subset of settings from configuration `settings`."""

    @property
    @abstractmethod
    def dataset(self) -> xr.Dataset:
        """The current dataset."""

    @property
    @abstractmethod
    def access_latency(self) -> float | None:
        """The time in seconds that it took for opening the dataset.
        `None` if the dataset has not been opened from `file_path`.
        """

    @abstractmethod
    def report(
        self,
        message: str,
        *,
        fatal: bool | None = None,
        suggestions: list[Suggestion | str] | None = None,
    ):
        """Report an issue.

        Args:
            message: mandatory message text
            fatal: True, if a fatal error is reported.
            suggestions: A list of suggestions for the user
                on how to fix the reported issue. Items may
                be of type `Suggestion` or `str`.
        """


class RuleExit(Exception):
    """The `RuleExit` is an exception that can be raised to
    immediately cancel dataset node validation with the current rule.

    Raise it from any of your `RuleOp` method implementations if further
    node traversal doesn't make sense. Typical usage:

    ```python
    if something_is_not_ok:
        ctx.report("Something is not ok.")
        raise RuleExit
    ```
    """


class RuleOp(ABC):
    """Define the specific rule validation operations."""

    def validate_datatree(self, ctx: RuleContext, node: DataTreeNode) -> None:
        """Validate the given datatree node.

        Args:
            ctx: The current rule context.
            node: The datatree node.

        Raises:
            RuleExit: to exit rule logic and further node traversal
        """

    def validate_dataset(self, ctx: RuleContext, node: DatasetNode) -> None:
        """Validate the given dataset node.

        Args:
            ctx: The current rule context.
            node: The dataset node.

        Raises:
            RuleExit: to exit rule logic and further node traversal
        """

    def validate_variable(self, ctx: RuleContext, node: VariableNode) -> None:
        """Validate the given data array (variable) node.

        Args:
            ctx: The current rule context.
            node: The data array (variable) node.

        Raises:
            RuleExit: to exit rule logic and further node traversal
        """

    def validate_attrs(self, ctx: RuleContext, node: AttrsNode) -> None:
        """Validate the given attributes node.

        Args:
            ctx: The current rule context.
            node: The attributes node.

        Raises:
            RuleExit: to exit rule logic and further node traversal
        """

    def validate_attr(self, ctx: RuleContext, node: AttrNode) -> None:
        """Validate the given attribute node.

        Args:
            ctx: The current rule context.
            node: The attribute node.

        Raises:
            RuleExit: to exit rule logic and further node traversal
        """


@dataclass(kw_only=True)
class RuleMeta(OperationMeta):
    """Rule metadata."""

    name: str
    """Rule name. Mandatory."""

    version: str = "0.0.0"
    """Rule version. Defaults to `0.0.0`."""

    description: str | None = None
    """Rule description."""

    docs_url: str | None = None
    """Rule documentation URL."""

    schema: dict[str, Any] | list[dict[str, Any]] | bool | None = None
    """JSON Schema used to specify and validate the rule operation
    options.

    It can take the following values:

    - Use `None` (the default) to indicate that the rule operation
      as no options at all.
    - Use a schema to indicate that the rule operation
      takes keyword arguments only.
      The schema's type must be `"object"`.
    - Use a list of schemas to indicate that the rule operation
      takes positional arguments only.
      If given, the number of schemas in the list specifies the
      number of positional arguments that must be configured.
    """

    type: Literal["problem", "suggestion", "layout"] = "problem"
    """Rule type. Defaults to `"problem"`.

    The type field can have one of the following values:

    - `"problem"`: Indicates that the rule addresses datasets that are
      likely to cause errors or unexpected behavior during runtime.
      These issues usually represent real bugs or potential runtime problems.
    - `"suggestion"`: Used for rules that suggest structural improvements
      or enforce best practices. These issues are not necessarily bugs, but
      following the suggestions may lead to more readable, maintainable, or
      consistent datasets.
    - `"layout"`: Specifies that the rule enforces consistent stylistic
      aspects of dataset formatting, e.g., whitespaces in names.
      Issues with layout rules are often automatically fixable
      (not supported yet).

    Primarily serves to categorize the rule's purpose for the benefit
    of developers and tools that consume XRLint output.
    It doesn’t directly affect the linting logic - that part is handled
    by the rule’s implementation and its configured severity.
    """

    @classmethod
    def value_name(cls) -> str:
        return "rule_meta"

    @classmethod
    def value_type_name(cls) -> str:
        return "RuleMeta | dict"


@dataclass(frozen=True)
class Rule(Operation):
    """A rule comprises rule metadata and a reference to the
    class that implements the rule's logic.

    Instances of this class can be easily created and added to a plugin
    by using the decorator `@define_rule` of the `Plugin` class.

    Args:
        meta: the rule's metadata
        op_class: the class that implements the rule's logic
    """

    meta: RuleMeta
    """Rule metadata of type `RuleMeta`."""

    op_class: Type[RuleOp]
    """The class the implements the rule's validation operation.
    The class must implement the `RuleOp` interface.
    """

    @classmethod
    def meta_class(cls) -> Type:
        return RuleMeta

    @classmethod
    def op_base_class(cls) -> Type:
        return RuleOp

    @classmethod
    def value_name(cls) -> str:
        return "rule"


@dataclass(frozen=True)
class RuleConfig(ValueConstructible, JsonSerializable):
    """A rule configuration.

    You should not use the class constructor directly.
    Instead, use its [from_value][xrlint.rule.RuleConfig.from_value]
    class method. The method's argument value can either be a
    rule _severity_, or a list where the first element is a rule
    _severity_ and subsequent elements are rule arguments:

    - _severity_
    - `[`_severity_`]`
    - `[`_severity_`,` _arg-1 | kwargs_ `]`
    - `[`_severity_`,` _arg-1_`,` _arg-2_`,` ...`,` _arg-n | kwargs_`]`

    The rule _severity_ is either

    - one of `"error"`, `"warn"`, `"off"` or
    - one of `2` (error), `1` (warn), `0` (off)

    Args:
        severity: rule severity, one of `2` (error), `1` (warn), or `0` (off)
        args: rule operation arguments.
        kwargs: rule operation keyword-arguments.
    """

    severity: Literal[0, 1, 2]
    """Rule severity, one of `2` (error), `1` (warn), or `0` (off)."""

    args: tuple[Any, ...] = field(default_factory=tuple)
    """Rule operation arguments."""

    kwargs: dict[str, Any] = field(default_factory=dict)
    """Rule operation keyword-arguments."""

    @classmethod
    def _convert_severity(cls, value: int | str) -> Literal[2, 1, 0]:
        try:
            # noinspection PyTypeChecker
            return SEVERITY_ENUM[value]
        except KeyError:
            raise ValueError(
                format_message_one_of("severity", value, SEVERITY_ENUM_TEXT)
            )

    @classmethod
    def _from_bool(cls, value: bool, name: str) -> "RuleConfig":
        return RuleConfig(cls._convert_severity(int(value)))

    @classmethod
    def _from_int(cls, value: int, name: str) -> "RuleConfig":
        return RuleConfig(cls._convert_severity(value))

    @classmethod
    def _from_str(cls, value: str, value_name: str) -> "RuleConfig":
        return RuleConfig(cls._convert_severity(value))

    @classmethod
    def _from_sequence(cls, value: Sequence, value_name: str) -> "RuleConfig":
        if not value:
            raise ValueError(f"{value_name} must not be empty")
        severity = cls._convert_severity(value[0])
        options = value[1:]
        if not options:
            args, kwargs = (), {}
        elif isinstance(options[-1], dict):
            args, kwargs = options[:-1], options[-1]
        else:
            args, kwargs = options, {}
        # noinspection PyTypeChecker
        return RuleConfig(severity, tuple(args), dict(kwargs))

    @classmethod
    def value_name(cls) -> str:
        return "rule_config"

    @classmethod
    def value_type_name(cls) -> str:
        return "int | str | list"

    # noinspection PyUnusedLocal
    def to_json(self, value_name: str | None = None) -> int | list:
        if not self.args and not self.kwargs:
            return self.severity
        else:
            return [self.severity, *self.args, self.kwargs]


def define_rule(
    name: str | None = None,
    version: str = "0.0.0",
    type: Literal["problem", "suggestion", "layout"] = "problem",
    description: str | None = None,
    docs_url: str | None = None,
    schema: dict[str, Any] | list[dict[str, Any]] | bool | None = None,
    registry: MutableMapping[str, Rule] | None = None,
    op_class: Type[RuleOp] | None = None,
) -> Callable[[Any], Type[RuleOp]] | Rule:
    """Define a rule.

    This function can be used to decorate your rule operation class
    definitions. When used as a decorator, the decorated operator class
    will receive a `meta` attribute of type [RuleMeta][xrlint.rule.RuleMeta].
    In addition, the `registry` if given, will be updated using `name`
    as key and a new [Rule][xrlint.rule.Rule] as value.

    Args:
        name: Rule name, see [RuleMeta][xrlint.rule.RuleMeta].
        version: Rule version, see [RuleMeta][xrlint.rule.RuleMeta].
        type: Rule type, see [RuleMeta][xrlint.rule.RuleMeta].
        description: Rule description,
            see [RuleMeta][xrlint.rule.RuleMeta].
        docs_url: Rule documentation URL,
            see [RuleMeta][xrlint.rule.RuleMeta].
        schema: Rule operation arguments schema,
            see [RuleMeta][xrlint.rule.RuleMeta].
        registry: Rule registry. Can be provided to register the
            defined rule using its `name`.
        op_class: Rule operation class. Must not be provided
            if this function is used as a class decorator.

    Returns:
        A decorator function, if `op_class` is `None` otherwise
            the value of `op_class`.

    Raises:
        TypeError: If either `op_class` or the decorated object is not
            a class derived from [RuleOp][xrlint.rule.RuleOp].
    """
    return Rule.define_operation(
        op_class,
        registry=registry,
        meta_kwargs=dict(
            name=name,
            version=version,
            description=description,
            docs_url=docs_url,
            type=type if type else "problem",
            schema=schema,
        ),
    )



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import xarray as xr

from xrlint.plugins.core.rules.var_desc import VarDesc
from xrlint.testing import RuleTest, RuleTester

pressure_attrs = dict(
    long_name="mean sea level pressure",
    units="hPa",
    standard_name="air_pressure_at_sea_level",
)

time_coord = xr.DataArray(
    [1, 2, 3], dims="time", attrs=dict(units="days since 2025-01-01")
)

valid_dataset_0 = xr.Dataset(
    coords=dict(time=time_coord),
)
valid_dataset_1 = xr.Dataset(
    data_vars=dict(pressure=xr.DataArray([1, 2, 3], dims="time", attrs=pressure_attrs)),
    coords=dict(time=time_coord),
)
valid_dataset_2 = xr.Dataset(
    data_vars=dict(
        chl=xr.DataArray(
            [1, 2, 3], dims="time", attrs=dict(description="It is air pressure")
        )
    ),
    coords=dict(time=time_coord),
)

invalid_dataset_0 = xr.Dataset(
    attrs=dict(),
    data_vars=dict(chl=xr.DataArray([1, 2, 3], dims="time", attrs=dict())),
    coords=dict(time=time_coord),
)

invalid_dataset_1 = xr.Dataset(
    attrs=dict(),
    data_vars=dict(
        chl=xr.DataArray(
            [1, 2, 3],
            dims="time",
            attrs=dict(standard_name="air_pressure_at_sea_level"),
        )
    ),
    coords=dict(time=time_coord),
)
invalid_dataset_2 = xr.Dataset(
    attrs=dict(),
    data_vars=dict(
        chl=xr.DataArray(
            [1, 2, 3], dims="time", attrs=dict(long_name="mean sea level pressure")
        )
    ),
    coords=dict(time=time_coord),
)
invalid_dataset_3 = xr.Dataset(
    attrs=dict(),
    data_vars=dict(chl=xr.DataArray([1, 2, 3], dims="time", attrs=pressure_attrs)),
    coords=dict(time=time_coord),
)

VarDescTest = RuleTester.define_test(
    "var-desc",
    VarDesc,
    valid=[
        RuleTest(dataset=valid_dataset_0),
        RuleTest(dataset=valid_dataset_1),
        RuleTest(dataset=valid_dataset_2, kwargs={"attrs": ["description"]}),
    ],
    invalid=[
        RuleTest(
            dataset=invalid_dataset_0,
            expected=[
                "Missing attribute 'standard_name'.",
                "Missing attribute 'long_name'.",
            ],
        ),
        RuleTest(
            dataset=invalid_dataset_1, expected=["Missing attribute 'long_name'."]
        ),
        RuleTest(
            dataset=invalid_dataset_2, expected=["Missing attribute 'standard_name'."]
        ),
        RuleTest(
            dataset=invalid_dataset_3,
            kwargs={"attrs": ["description"]},
            expected=["Missing attribute 'description'."],
        ),
    ],
)



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

import re

from xrlint.node import VariableNode
from xrlint.plugins.core.plugin import plugin
from xrlint.rule import RuleContext, RuleOp

_EXAMPLE_UNIT_1 = "seconds since 2010-10-8 15:15:42.5 -6:00"
_EXAMPLE_UNIT_2 = "days since 2000-01-01 +0:00"

_AMBIGUOUS_UNITS_OF_TIME = (
    "years",
    "year",
    "y",
    "months",
    "month",
    "m",
)

_UNAMBIGUOUS_UNITS_OF_TIME = (
    "days",
    "day",
    "d",
    "hours",
    "hour",
    "hr",
    "h",
    "minutes",
    "minute",
    "min",
    "seconds",
    "second",
    "sec",
    "s",
)

_ALL_UNITS_OF_TIME = (*_AMBIGUOUS_UNITS_OF_TIME, *_UNAMBIGUOUS_UNITS_OF_TIME)

_RE_DATE = re.compile(r"^\d{4}-\d{1,2}-\d{1,2}$")
_RE_TIME = re.compile(r"^\d{1,2}:\d{1,2}:\d{1,2}(\.\d{1,6})?$")
_RE_TZ = re.compile(r"^[+-]\d{1,2}:\d{1,2}$")


@plugin.define_rule(
    "time-coordinate",
    version="1.0.0",
    type="problem",
    description=(
        "Time coordinates should have valid and unambiguous time units encoding."
    ),
    docs_url=(
        "https://cfconventions.org/cf-conventions/cf-conventions.html#time-coordinate"
    ),
)
class TimeCoordinate(RuleOp):
    def validate_variable(self, ctx: RuleContext, node: VariableNode):
        array = node.array
        attrs = array.attrs
        encoding = array.encoding
        attrs = array.attrs

        units: str | None = None
        source: str | None = None
        if "units" in encoding:
            units = encoding["units"]
            source = "encoding"
        elif "units" in attrs:
            units = attrs["units"]
            source = "attribute"

        if source is None:
            if _is_time_by_name(attrs):
                ctx.report("Missing encoding/attribute 'units'.")
            # No more to check w.o. time units
            return
        elif not isinstance(units, str):
            if _is_time_by_name(attrs):
                ctx.report(
                    f"Invalid {source} 'units',"
                    f" expected type str, got {type(units).__name__}."
                )
            # No more to check w.o. time units
            return

        units_ok = True
        units_parts = units.split(" ")
        num_unit_parts = len(units_parts)
        is_time_by_units = num_unit_parts >= 3 and units_parts[1] == "since"
        if not is_time_by_units:
            if not _is_time_by_name(attrs):
                # Not a time coordinate
                return
            units_ok = False
        else:
            # We have time units
            calendar: str | None = None
            if source == "encoding" and "calendar" in encoding:
                calendar = encoding["calendar"]
            elif source == "attribute" and "calendar" in attrs:
                calendar = attrs["calendar"]
            # Note, we should also check here for valid calendar names
            if calendar is None:
                ctx.report(f"Missing {source} 'calendar'.")

            uot_part = units_parts[0]
            date_part = units_parts[2]
            time_part = None
            tz_part = None

            if uot_part not in _ALL_UNITS_OF_TIME:
                ctx.report(
                    f"Unrecognized units of measure for time"
                    f" in {source} 'units': {units!r}.",
                    suggestions=[
                        _units_format_suggestion(),
                        _units_of_time_suggestion(),
                    ],
                )
            elif uot_part in _AMBIGUOUS_UNITS_OF_TIME:
                ctx.report(
                    f"Ambiguous units of measure for time in"
                    f" {source} 'units': {units!r}.",
                    suggestions=[
                        _units_format_suggestion(),
                        _units_of_time_suggestion(),
                    ],
                )

            if num_unit_parts == 3:
                pass
            elif num_unit_parts == 4:
                time_or_tz_part = units_parts[3]
                if _RE_TIME.match(time_or_tz_part):
                    time_part = time_or_tz_part
                else:
                    tz_part = time_or_tz_part
            elif num_unit_parts == 5:
                time_part = units_parts[3]
                tz_part = units_parts[4]
            else:
                time_part = units_parts[-2]
                tz_part = units_parts[-1]
                units_ok = False

            if units_ok and not _RE_DATE.match(date_part):
                units_ok = False
            if units_ok and time_part and not _RE_TIME.match(time_part):
                units_ok = False
            if units_ok and tz_part and not _RE_TZ.match(tz_part):
                units_ok = False

            if not tz_part:
                ctx.report(
                    f"Missing timezone in {source} 'units': {units!r}.",
                    suggestions=[
                        _units_format_suggestion(),
                        f"Append timezone specification, e.g., use"
                        f" {' '.join(units_parts[:-1] + ['+0:00'])!r}.",
                    ],
                )

        if not units_ok:
            ctx.report(
                f"Invalid {source} 'units': {units!r}.",
                suggestions=[_units_format_suggestion()],
            )


def _units_format_suggestion():
    use_units_format_msg = (
        f"Specify units using the UDUNITS format,"
        f" e.g., {_EXAMPLE_UNIT_1!r} or {_EXAMPLE_UNIT_2!r}."
    )
    return use_units_format_msg


def _units_of_time_suggestion():
    return f"Use one of {', '.join(map(repr, _UNAMBIGUOUS_UNITS_OF_TIME))}"


def _is_time_by_name(attrs):
    return attrs.get("standard_name") == "time" or attrs.get("axis") == "T"



#  Copyright © 2025 Brockmann Consult GmbH.
#  This software is distributed under the terms and conditions of the
#  MIT license (https://mit-license.org/).

from unittest import TestCase

import pytest

from xrlint.plugin import Plugin
from xrlint.util.importutil import ValueImportError, import_submodules, import_value


def get_foo():
    return 42


class ImportSubmodulesTest(TestCase):
    def test_import_submodules(self):
        modules = import_submodules("tests.util.test_importutil_pkg", dry_run=True)
        self.assertEqual(
            {
                "tests.util.test_importutil_pkg.module1",
                "tests.util.test_importutil_pkg.module2",
            },
            set(modules),
        )

        modules = import_submodules("tests.util.test_importutil_pkg")
        self.assertEqual(
            {
                "tests.util.test_importutil_pkg.module1",
                "tests.util.test_importutil_pkg.module2",
            },
            set(modules),
        )


class ImportValueTest(TestCase):
    def test_import_exported_value_ok(self):
        plugin, plugin_ref = import_value(
            "xrlint.plugins.core", "export_plugin", factory=Plugin.from_value
        )
        self.assertIsInstance(plugin, Plugin)
        self.assertEqual("xrlint.plugins.core:export_plugin", plugin_ref)

    def test_import_exported_value_ok_no_factory(self):
        value, value_ref = import_value(
            "tests.util.test_importutil:get_foo",
        )
        self.assertEqual(value, 42)
        self.assertEqual("tests.util.test_importutil:get_foo", value_ref)

    # noinspection PyMethodMayBeStatic
    def test_import_exported_value_fail(self):
        with pytest.raises(
            ValueImportError,
            match=(
                r"value of tests.util.test_importutil:get_foo\(\)"
                r" must be of type float, but got int"
            ),
        ):
            import_value("tests.util.test_importutil:get_foo", expected_type=float)

    # noinspection PyMethodMayBeStatic
    def test_import_exported_value_import_error(self):
        with pytest.raises(
            ValueImportError,
            match=(
                "failed to import value from 'tests.util.test_baz:get_foo':"
                " No module named 'tests.util.test_baz'"
            ),
        ):
            import_value("tests.util.test_baz:get_foo")
